import pytest

from easypy.bunch import Bunch

import easypy.typed_struct as ts


def test_typed_struct_validations():
    class OutOfRangeError(Exception):
        pass

    class Foo(ts.TypedStruct):
        a = ts.Field(int, default=20)
        a.add_validation(range(10, 30).__contains__, OutOfRangeError)
        b = ts.Field(str)

    # Valid construction
    Foo(b='hi')
    Foo(a=14, b='hello')

    # Invalid construction
    with pytest.raises(ts.NotFields) as exc:
        Foo(b='', c=10)
    assert exc.value.message == 'c are not fields of Foo'

    with pytest.raises(ts.MissingField) as exc:
        Foo()
    assert exc.value.message == 'Foo is missing mandatory field b'

    with pytest.raises(ts.FieldTypeMismatch):
        Foo(a=20.0, b='')

    with pytest.raises(OutOfRangeError):
        Foo(a=100, b='')

    foo = Foo(b='')

    # Valid attribute assignments
    foo.a = 15
    foo.b = 'hello'

    # Invalid attribute assignments
    with pytest.raises(ts.NotAField) as exc:
        foo.c = 10
    assert exc.value.message == 'c is not a field of Foo'

    with pytest.raises(ts.FieldTypeMismatch):
        foo.b = []

    with pytest.raises(OutOfRangeError):
        foo.a = 1000

    assert foo.a == 15
    assert foo.b == 'hello'

    # Valid key assignments
    foo['a'] = 25
    foo['b'] = 'hi'

    # Invalid key assignments
    with pytest.raises(ts.NotAKey) as exc:
        foo['c'] = 10
    assert exc.value.message == 'c is not a key of Foo'

    with pytest.raises(ts.FieldTypeMismatch):
        foo['b'] = []

    with pytest.raises(OutOfRangeError):
        foo['a'] = 1000

    assert foo['a'] == 25
    assert foo['b'] == 'hi'


def test_typed_struct_nesting():
    class Foo(ts.TypedStruct):
        a = ts.Field(int, default=0)
        b = ts.Field(int, default=1)

    class Bar(ts.TypedStruct):
        c = ts.Field(int, default=2)
        d = ts.Field(int)

    class Baz(ts.TypedStruct):
        foo = ts.Field(Foo, default=Foo(a=10))
        foo2 = Foo
        bar = Bar

    baz = Baz(bar=dict(d=3))
    assert baz.foo == Foo(a=10, b=1)
    assert baz.foo2 == Foo()
    assert baz.bar == Bar(c=2, d=3)

    # Very important - nested fields are modifiable so if the default is used
    # they must not be shared!
    assert baz.foo is baz.foo
    assert baz.foo is not Baz(bar=dict(d=3)).foo


def test_typed_struct_comparison():
    class Foo(ts.TypedStruct):
        a = int
        b = int

    class Bar(ts.TypedStruct):
        a = int
        b = int

    assert Foo(a=1, b=2) == Foo(a=1, b=2)
    assert Foo(a=1, b=2) != Foo(a=1, b=3)
    assert Foo(a=1, b=2) != Bar(a=1, b=2)

    assert Foo(a=1, b=2).to_dict() == Bar(a=1, b=2).to_dict()
    assert Foo(a=1, b=2).to_bunch() == Bar(a=1, b=2).to_bunch()


@pytest.mark.parametrize("use_full_syntax", [True, False])
def test_typed_struct_list_fields(use_full_syntax):
    class Foo(ts.TypedStruct):
        a = int

    if use_full_syntax:
        class Bar(ts.TypedStruct):
            foos = ts.Field([Foo])
            nums = ts.Field([int])
    else:
        class Bar(ts.TypedStruct):
            foos = [Foo]
            nums = [int]

    # C'tor
    bar = Bar(foos=[Foo(a=1), Foo(a=2)], nums=[1, 2, 3])
    assert bar.foos == [Foo(a=1), Foo(a=2)]
    assert bar.nums == [1, 2, 3]

    # Default
    bar = Bar()
    assert bar.foos == []
    assert bar.nums == []

    # append()
    bar.foos.append(Foo(a=1))
    bar.foos.append(dict(a=2))
    with pytest.raises(ts.FieldTypeMismatch):
        bar.foos.append(3)
    assert bar.foos == [Foo(a=1), Foo(a=2)]

    bar.nums.append(1)
    with pytest.raises(ts.FieldTypeMismatch):
        bar.nums.append('2')
    assert bar.nums == [1]

    # __setitem__()
    bar.foos[0] = Foo(a=4)
    bar.foos[1] = dict(a=5)
    with pytest.raises(ts.FieldTypeMismatch):
        bar.foos[0] = 6
    assert bar.foos == [Foo(a=4), Foo(a=5)]

    bar.nums[0] = 3
    assert bar.nums == [3]

    # insert()
    bar.foos.insert(0, Foo(a=7))
    bar.foos.insert(2, dict(a=8))
    with pytest.raises(ts.FieldTypeMismatch):
        bar.foos.insert(1, 9)
    assert bar.foos == [Foo(a=7), Foo(a=4), Foo(a=8), Foo(a=5)]

    bar.nums.insert(0, 4)
    assert bar.nums == [4, 3]

    # Assignment
    bar.foos = [Foo(a=10), dict(a=11)]
    with pytest.raises(ts.FieldTypeMismatch):
        bar.foos = [12]
    assert bar.foos == [Foo(a=10), Foo(a=11)]

    bar.nums = [5]
    assert bar.nums == [5]

    # __iadd__()
    bar.foos += [Foo(a=13), dict(a=14)]
    with pytest.raises(ts.FieldTypeMismatch):
        bar.foos += [15]
    assert bar.foos == [Foo(a=10), Foo(a=11), Foo(a=13), Foo(a=14)]

    bar.nums += [6]
    assert bar.nums == [5, 6]

    # extend()
    bar.foos.extend(iter([Foo(a=16), dict(a=17)]))
    with pytest.raises(ts.FieldTypeMismatch):
        bar.foos.extend(range(18))
    assert bar.foos == [Foo(a=10), Foo(a=11), Foo(a=13), Foo(a=14), Foo(a=16), Foo(a=17)]

    bar.nums.extend(iter([7]))
    assert bar.nums == [5, 6, 7]

    # Make sure that bar.foos and bar.nums are still a ListCollection
    assert bar.foos.M.a.C == [10, 11, 13, 14, 16, 17]
    assert bar.nums.M.call(lambda x: x * 10).C == [50, 60, 70]


@pytest.mark.parametrize("use_full_syntax", [True, False])
def test_typed_struct_dict_fields(use_full_syntax):
    class Foo(ts.TypedStruct):
        a = int

    if use_full_syntax:
        class Bar(ts.TypedStruct):
            foos = ts.Field({int: Foo})
            nums = ts.Field({int: int})
    else:
        class Bar(ts.TypedStruct):
            foos = {int: Foo}
            nums = {int: int}

    # C'tor
    bar = Bar(foos={1: Foo(a=2)}, nums={3: 4})
    assert bar.foos == {1: Foo(a=2)}
    assert bar.nums == {3: 4}

    # Default
    bar = Bar()
    assert bar.foos == {}
    assert bar.nums == {}

    # __setitem__()

    bar.foos[1] = Foo(a=2)
    bar.foos[3] = dict(a=4)
    with pytest.raises(ts.FieldTypeMismatch):
        bar.foos[6] = 7
    with pytest.raises(ts.FieldKeyTypeMismatch):
        bar.foos['8'] = Foo(a=9)
    assert bar.foos == {1: Foo(a=2), 3: Foo(a=4)}

    bar.nums[1] = 2
    with pytest.raises(ts.FieldTypeMismatch):
        bar.nums[3] = '4'
    with pytest.raises(ts.FieldKeyTypeMismatch):
        bar.nums['5'] = 6
    assert bar.nums == {1: 2}

    # Assignment

    bar.foos = {5: dict(a=6), 7: Foo(a=8)}
    with pytest.raises(ts.FieldTypeMismatch):
        bar.foos = {9: 10}
    with pytest.raises(ts.FieldKeyTypeMismatch):
        bar.foos = {'11': Foo(a=12)}
    assert bar.foos == {5: Foo(a=6), 7: Foo(a=8)}

    bar.nums = {7: 8}
    with pytest.raises(ts.FieldTypeMismatch):
        bar.nums = {9: '10'}
    with pytest.raises(ts.FieldKeyTypeMismatch):
        bar.nums = {'11': 12}
    assert bar.nums == {7: 8}

    # setdefault()

    assert bar.foos.setdefault(5, Foo(a=13)).a == 6  # from the assignments tests
    assert bar.foos.setdefault(14, Foo(a=15)).a == 15
    assert bar.foos.setdefault(16, dict(a=17)).a == 17
    with pytest.raises(ts.FieldTypeMismatch):
        bar.foos.setdefault(18, 19)
    with pytest.raises(ts.FieldKeyTypeMismatch):
        bar.foos.setdefault('20', Foo(a=21))
    assert bar.foos == {5: Foo(a=6), 7: Foo(a=8), 14: Foo(a=15), 16: Foo(a=17)}

    assert bar.nums.setdefault(7, 13) == 8  # NOTE: from the assignments tests
    assert bar.nums.setdefault(14, 15) == 15
    with pytest.raises(ts.FieldTypeMismatch):
        bar.nums.setdefault(16, '17')
    with pytest.raises(ts.FieldKeyTypeMismatch):
        bar.nums.setdefault('18', 19)
    assert bar.nums == {7: 8, 14: 15}

    bar.foos.clear()  # it got too big

    # update()

    # NOTE: I can't test the .update(bla=...) syntax, because I only accept
    # ints here.

    bar.foos.update({22: Foo(a=23), 24: dict(a=25)})
    with pytest.raises(ts.FieldTypeMismatch):
        bar.foos.update({26: 27})
    with pytest.raises(ts.FieldKeyTypeMismatch):
        bar.foos.update({'28': Foo(a=29)})
    assert bar.foos == {22: Foo(a=23), 24: Foo(a=25)}

    bar.nums.update({20: 21})
    with pytest.raises(ts.FieldTypeMismatch):
        bar.nums.update({22: '23'})
    with pytest.raises(ts.FieldKeyTypeMismatch):
        bar.nums.update({'24': 25})
    assert bar.nums == {7: 8, 14: 15, 20: 21}


@pytest.mark.parametrize("use_full_syntax", [True, False])
def test_typed_struct_bunch_fields(use_full_syntax):
    if use_full_syntax:
        class Bar(ts.TypedStruct):
            nums = ts.Field({str: int})
    else:
        class Bar(ts.TypedStruct):
            nums = {str: int}

    # NOTE: TypedBunch shares most things with TypedDict, so I'm only doing a quick sanity test

    bar = Bar(nums=Bunch(a=0))
    assert bar.nums.a == 0

    bar.nums.a = 1
    bar.nums['b'] = 2
    bar.nums.update(c=3)
    with pytest.raises(ts.FieldTypeMismatch):
        bar.nums.d = '4'
    with pytest.raises(ts.FieldKeyTypeMismatch):
        bar.nums[5] = 6

    assert bar.nums == Bunch(a=1, b=2, c=3)


def test_typed_struct_field_defaults():
    class Foo(ts.TypedStruct):
        x = int

    class Bar(ts.TypedStruct):
        a = ts.Field(int, default=1)
        b = ts.Field([int], default=[2, 3])
        c = ts.Field(Foo, default=Foo(x=4))
        d = ts.Field(Foo, default=dict(x=5))
        e = ts.Field([Foo], default=[Foo(x=6), dict(x=7)])
        f = ts.Field({str: Foo}, default=dict(g=Foo(x=8)))

    bar1 = Bar()
    bar2 = Bar()

    assert bar1.a == bar2.a == 1

    assert bar1.b == bar2.b == [2, 3]
    assert bar1.b is not bar2.b

    assert bar1.c == bar2.c == Foo(x=4)
    assert bar1.c is not bar2.c

    assert bar1.d == bar2.d == Foo(x=5)
    assert bar1.d is not bar2.d

    assert bar1.e == bar2.e == [Foo(x=6), Foo(x=7)]
    assert bar1.e is not bar2.e
    assert bar1.e[0] is not bar2.e[0]
    assert bar1.e[1] is not bar2.e[1]

    assert bar1.f == bar2.f == Bunch(g=Foo(x=8))
    assert bar1.f is not bar2.f
    assert bar1.f.g is not bar2.f.g


def test_typed_struct_metadata():
    class Foo(ts.TypedStruct):
        a = ts.Field(int, meta=dict(format='%d'))
        b = ts.Field(int, meta=dict(format='%03d'))

    class Bar(ts.TypedStruct):
        c = ts.Field(str, meta=dict(format='%s'))
        d = ts.Field(str, meta=dict(format='"%s"'))

    def format_object(obj):
        def gen():
            for k, v in obj.items():
                field = getattr(type(obj), k)
                formatted = field.meta.format % (v,)
                yield '%s=%s' % (k, formatted)
        return '%s[%s]' % (type(obj).__name__, ', '.join(gen()))

    assert format_object(Foo(a=1, b=2)) == 'Foo[a=1, b=002]'
    assert format_object(Bar(c='3', d='4')) == 'Bar[c=3, d="4"]'


def test_typed_struct_convert():
    class Foo(ts.TypedStruct):
        a = ts.Field(int)
        a.add_conversion(str, int)
        a.add_conversion(float, int)

    assert Foo(a=1).a == 1
    assert Foo(a='2').a == 2
    assert Foo(a=3.4).a == 3
    with pytest.raises(ts.FieldTypeMismatch):
        # Don't know how to convert complex
        Foo(a=5 + 6j).a


@pytest.mark.parametrize("use_helper_methods", [True, False])
def test_typed_struct_preprocess(use_helper_methods):
    class Error1(Exception):
        pass

    class Error2(Exception):
        pass

    if use_helper_methods:
        class Foo(ts.TypedStruct):
            a = ts.Field(int)

            a.add_validation(lambda n: n % 2 == 0, Error1, 'not an even number')
            a.add_validation(range(10, 31).__contains__, Error2, 'out of range([10, 30])')

            a.add_conversion(str, int)
            a.add_conversion(float, int)
    else:
        def preprocess(obj):
            if isinstance(obj, str) or isinstance(obj, float):
                obj = int(obj)

            if isinstance(obj, int):
                if obj % 2 == 1:
                    raise Error1('not an even number')
                if obj not in range(10, 31):
                    raise Error2('out of range([10, 30])')

            return obj

        class Foo(ts.TypedStruct):
            a = ts.Field(int, preprocess=preprocess)

    assert Foo(a=12).a == 12
    assert Foo(a='12').a == 12
    assert Foo(a=12.0).a == 12

    with pytest.raises(ts.FieldTypeMismatch):
        Foo(a=[])

    # Raising errors on validations:

    with pytest.raises(Error1) as exc:
        Foo(a=13)
    assert exc.value.args[0] == 'not an even number'

    with pytest.raises(Error2) as exc:
        Foo(a=32)
    assert exc.value.args[0] == 'out of range([10, 30])'

    # Raising errors on validations even after conversions:

    with pytest.raises(Error1):
        Foo(a=13.0)

    with pytest.raises(Error2):
        Foo(a='32')


def test_typed_struct_collection_type_verification():
    """
    To prevent things like this:

        In [1]: from easypy import typed_struct as ts

        In [2]: class Foo(ts.TypedStruct):
           ...:     bars = [str]
           ...:

        In [3]: Foo(bars='hello')
        Out[3]:
        Foo(bars=['h',
                  'e',
                  'l',
                  'l',
                  'o'])
    """

    class Foo(ts.TypedStruct):
        a = ts.Field([str])
        b = ts.Field({str: str})
        c = ts.Field({int: str})

    # Test list:

    assert Foo(a=['1']).a == ['1']
    with pytest.raises(ts.FieldCollectionTypeMismatch):
        Foo(a='2')
    with pytest.raises(ts.FieldCollectionTypeMismatch):
        Foo(a={'3': '4'})

    # Test dict:

    assert Foo(b={'5': '6'}).b == {'5': '6'}
    with pytest.raises(ts.FieldCollectionTypeMismatch):
        Foo(b='8')
    with pytest.raises(ts.FieldCollectionTypeMismatch):
        Foo(b=['9'])

    # Test Bunch:

    assert Foo(c={10: '11'}).c == {10: '11'}
    with pytest.raises(ts.FieldCollectionTypeMismatch):
        Foo(c='13')
    with pytest.raises(ts.FieldCollectionTypeMismatch):
        Foo(c=['14'])


def test_typed_struct_auto_field_wrapping_dsl():
    class Foo(ts.TypedStruct):
        a = int
        a.default = 1
        b = float
        b.convertible_from(str)

    assert Foo(b='2.3').to_dict() == dict(a=1, b=2.3)


def test_typed_struct_inheritance():
    class Foo(ts.TypedStruct):
        a = int
        a.default = 1

        b = int
        b.default = 2

    class Bar(Foo):
        a.default = 3

        c = int
        c.default = 4

    assert Foo().to_dict() == dict(a=1, b=2)
    assert Bar().to_dict() == dict(a=3, b=2, c=4)


def test_typed_struct_repr():
    class Foo(ts.TypedStruct):
        a = int
        b = float
        b.repr = False
        c = str
        c.repr = '`{}`'.format

    assert repr(Foo(a=1, b=2.0, c='3')) == 'Foo(a=1, c=`3`)'


def test_typed_struct_hash():
    class Foo(ts.TypedStruct):
        a = int
        b = int
        b.hash = False
        c = int
        c.hash = lambda n: n % 2


    assert hash(Foo(a=1, b=2, c=3)) == hash(Foo(a=1, b=2, c=3))
    assert hash(Foo(a=1, b=2, c=3)) != hash(Foo(a=2, b=2, c=3))
    assert hash(Foo(a=1, b=2, c=3)) == hash(Foo(a=1, b=1, c=3)), 'b should not be part of the hash'
    assert hash(Foo(a=1, b=2, c=3)) == hash(Foo(a=1, b=2, c=1)), 'Only the parity of c should be in the hash'
    assert hash(Foo(a=1, b=2, c=3)) != hash(Foo(a=1, b=2, c=2)), 'c with different parity should generate different hash'

    class Bar(ts.TypedStruct):
        a = int
        b = int
        b.hash = False
        c = int
        c.hash = lambda n: n % 2

    assert hash(Foo(a=1, b=2, c=3)) != hash(Bar(a=1, b=2, c=3)), \
        'different classes with same fields should have different hahses'



from __future__ import absolute_import
import sys
import traceback
from time import time
from datetime import datetime
from contextlib import contextmanager
from textwrap import indent
from logging import getLogger
_logger = getLogger(__name__)


class PException(Exception):

    """An exception object that can accept kwargs as attributes"""

    def __init__(self, message="", *args, **params):
        if args or params:
            message = message.format(*args, **params)
        Exception.__init__(self, message)
        self.context = params.pop("context", None)
        self.traceback = params.pop("traceback", None)
        if self.traceback is True:
            self.traceback = traceback.format_exc()
        self.message = message
        self.timestamp = params.pop('timestamp', time())
        if 'tip' not in params:
            # sometimes it's on the class
            params['tip'] = getattr(self, 'tip', None)
        self._params = {}
        self.add_params(**params)

    def __reduce__(self):
        return (self.__class__.__new__, (self.__class__,), self.__getstate__())

    def __getstate__(self):
        return (self.message, self.context, self.traceback, self.timestamp, self._params)

    def __setstate__(self, state):
        self.message, self.context, self.traceback, self.timestamp, params = state
        self._params = {}
        self.add_params(**params)

    def add_params(self, **params):
        for k, v in params.items():
            setattr(self, k, v)
        self._params.update(params)

    def __repr__(self):
        if self._params:
            kw = sorted("%s=%r" % (k, v) for k, v in self._params.items())
            return "%s(%r, %s)" % (self.__class__.__name__, self.message, ", ".join(kw))
        else:
            return "%s(%r)" % (self.__class__.__name__, self.message)

    def __str__(self):
        return self.render(traceback=False, color=False)

    def render(self, params=True, context=True, traceback=True, timestamp=True, color=True):
        text = ""

        if self.message:
            text += "".join("WHITE<<%s>>\n" % line for line in self.message.splitlines())

        if params and self._params:
            tip = self._params.pop('tip', None)
            text += indent("".join(make_block(self._params)), " " * 4)
            if tip:
                tip = tip.format(**self._params)
                lines = tip.splitlines()
                text += indent("GREEN(BLUE)@{tip = %s}@\n" % lines[0], " " * 4)
                for line in lines[1:]:
                    text += indent("GREEN(BLUE)@{      %s}@\n" % lines[0], " " * 4)
                self._params['tip'] = tip  # put it back in params, even though it might've been on the class

        if timestamp and self.timestamp:
            ts = datetime.fromtimestamp(self.timestamp).isoformat()
            text += indent("MAGENTA<<timestamp = %s>>\n" % ts, " " * 4)

        if context and self.context:
            text += "Context:\n" + indent("".join(make_block(self.context, skip={"indentation"})), " " * 4)

        if traceback and self.traceback:
            fmt = "DARK_GRAY@{{{}}}@"
            text += "\n".join(map(fmt.format, self.traceback.splitlines()))

        if not color:
            from easypy.colors import uncolored
            text = uncolored(text)

        return text

    @classmethod
    def make(cls, name):
        return type(name, (cls,), {})

    @classmethod
    @contextmanager
    def on_exception(cls, acceptable=Exception, **kwargs):
        try:
            yield
        except cls:
            # don't mess with exceptions of this type
            raise
        except acceptable as exc:
            exc_info = sys.exc_info()
            _logger.debug("'%s' raised; Raising as '%s'" % (type(exc), cls), exc_info=exc_info)
            raise cls(traceback=True, **kwargs) from None


def make_block(d, skip={}):
    for k in sorted(d):
        if k.startswith("_"):
            continue
        if k in skip:
            continue
        v = d[k]
        if isinstance(v, datetime):
            v = v.isoformat()
        elif not isinstance(v, str):
            v = repr(v)
        dark = False
        if k.startswith("~"):
            k = k[1:]
            dark = True
        head = "%s = " % k
        block = indent(v, " " * len(head))
        block = head + block[len(head):]
        if dark:
            block = "DARK_GRAY@{%s}@" % block
        yield block + "\n"


class TException(PException):

    @property
    def template(self):
        raise NotImplementedError("Must implement template")

    def __init__(self, *args, **params):
        super(TException, self).__init__(self.template, *args, **params)

    @classmethod
    def make(cls, name, template):
        return type(name, (cls,), dict(template=template))


def convert_traceback_to_list(tb):
    # convert to list of dictionaries that contain file, line_no and function
    traceback_list = [dict(file=file, line_no=line_no, function=function)
                      for file, line_no, function, _ in traceback.extract_tb(tb)]
    return traceback_list


def apply_timestamp(exc, now=None):
    timestamp = now or time()
    if getattr(exc, "timestamp", "__missing__") == "__missing__":
        try:
            exc.timestamp = timestamp
        except Exception:
            pass
    return exc



from easypy.timing import Timer, TimeInterval


def test_time_interval1():

    st = 150000000
    t = Timer(st)
    t.t1 = t.t0 + 1

    ti = t.to_interval()

    assert t in ti

    assert t.t0 in ti
    assert t.t1 in ti

    assert t.t0 - 1 not in ti
    assert t.t1 + 1 not in ti

    assert ti.duration == t.duration
    assert ti.duration == 1
    assert ti.duration_delta.total_seconds() == 1

    assert str(ti) == '<TI 05:40:00..(1s)..05:40:01>'

    assert str(t.to_interval().to_timer()) == str(t)


def test_time_interval2():
    st = 150000000
    ti = TimeInterval()
    assert str(ti) == '<TI Eternity>'

    ti = TimeInterval(from_time=st)
    assert str(ti) == '<TI 05:40:00...>'

    ti = TimeInterval(from_time=st, to_time=st)
    assert str(ti) == '<TI 05:40:00..(0.0ms)..05:40:00>'

    ti = TimeInterval(to_time=st)
    assert str(ti) == '<TI ...05:40:00>'



"""
This module is about synchronizing and coordinating events among concurrent activities.
"""


from contextlib import contextmanager, ExitStack
import sys
import threading
import time
import inspect
from functools import wraps
import re
import logging
import atexit
import signal
import os
from collections import Counter

import easypy._multithreading_init  # noqa
from .bunch import Bunch
from .gevent import is_module_patched
from .decorations import wrapper_decorator, parametrizeable_decorator
from .caching import locking_cache
from .exceptions import PException, TException
from .units import NEVER, MINUTE, HOUR
from .misc import Hex
from .humanize import time_duration  # due to interference with jrpc
from .misc import kwargs_resilient

_logger = logging.getLogger(__name__)
_verbose_logger = logging.getLogger('%s.locks' % __name__)  # logger for less important logs of RWLock.

IS_A_TTY = sys.stdout.isatty()


class TimeoutException(PException, TimeoutError):
    pass


class PredicateNotSatisfied(TimeoutException):
    # use this exception with 'wait', to indicate predicate is not satisfied
    # and allow it to raise a more informative exception
    pass


class LockLeaseExpired(TException):
    template = "Lock Lease Expired - thread is holding this lock for too long"


class ProcessExiting(TException):
    template = "Aborting thread - process is exiting"


class TimebombExpired(TException):
    template = "Timebomb Expired - process killed itself"
    exit_with_code = 234


_exiting = False  # we use this to break out of lock-acquisition loops


@atexit.register
def break_locks():
    global _exiting
    _exiting = True


def _check_exiting():
    if _exiting:
        raise ProcessExiting()


class TerminationSignal(TException):
    template = "Process got a termination signal: {_signal}"


class NotMainThread(TException):
    template = "Binding must be invoked from main thread"


class SignalAlreadyBound(TException):
    template = "Signal already bound to another signal handler(s)"


class LastErrorEmpty(TException):
    template = "Signal caught, but no error to raise"


LAST_ERROR = None
REGISTERED_SIGNAL = None


class NotInitialized(TException):
    template = "Signal type not initialized, must use bind_to_subthread_exceptions in the main thread"


def async_raise_in_main_thread(exc, use_concurrent_loop=True):
    """
    Uses a unix signal to raise an exception to be raised in the main thread.
    """

    from plumbum import local
    pid = os.getpid()
    if not REGISTERED_SIGNAL:
        raise NotInitialized()

    # sometimes the signal isn't caught by the main-thread, so we should try a few times (WEKAPP-14543)
    def do_signal(raised_exc):
        global LAST_ERROR
        if LAST_ERROR is not raised_exc:
            _logger.debug("MainThread took the exception - we're done here")
            if use_concurrent_loop:
                raiser.stop()
            return

        _logger.info("Raising %s in main thread", type(LAST_ERROR))
        local.cmd.kill("-%d" % REGISTERED_SIGNAL, pid)

    if use_concurrent_loop:
        from .concurrency import concurrent
        raiser = concurrent(do_signal, raised_exc=exc, loop=True, sleep=30, daemon=True, throw=False)
        raiser.start()
    else:
        do_signal(exc)



if is_module_patched("threading"):
    import gevent
    def _rimt(exc):
        _logger.info('YELLOW<<killing main thread greenlet>>')
        main_thread_greenlet = threading.main_thread()._greenlet
        orig_throw = main_thread_greenlet.throw

        # we must override "throw" method so exception will be raised with the original traceback
        def throw(*args):
            if len(args) == 1:
                ex = args[0]
                return orig_throw(ex.__class__, ex, ex.__traceback__)
            return orig_throw(*args)
        main_thread_greenlet.throw = throw
        gevent.kill(main_thread_greenlet, exc)
        _logger.debug('exiting the thread that failed')
        raise exc
else:
    _rimt = async_raise_in_main_thread


# must be invoked in main thread in "geventless" runs in order for raise_in_main_thread to work
def initialize_exception_listener():
    global REGISTERED_SIGNAL
    if REGISTERED_SIGNAL:
        # already registered
        return

    if threading.current_thread() is not threading.main_thread():
        raise NotMainThread()

    def handle_signal(sig, stack):
        global LAST_ERROR
        error = LAST_ERROR
        LAST_ERROR = None
        if error:
            raise error
        raise LastErrorEmpty(signal=sig)

    custom_signal = signal.SIGUSR1
    if signal.getsignal(custom_signal) in (signal.SIG_DFL, signal.SIG_IGN):  # check if signal is already trapped
        signal.signal(custom_signal, handle_signal)
        REGISTERED_SIGNAL = custom_signal
    else:
        raise SignalAlreadyBound(signal=custom_signal)


@contextmanager
def raise_in_main_thread(exception_type=Exception):

    try:
        yield
    except ProcessExiting:
        # this exception is meant to stay within the thread
        raise
    except exception_type as exc:
        if threading.current_thread() is threading.main_thread():
            raise
        exc._raised_asynchronously = True

        global LAST_ERROR
        if LAST_ERROR:
            _logger.warning("a different error (%s) is pending - skipping", type(LAST_ERROR))
            raise
        LAST_ERROR = exc
        _rimt(exc)


def initialize_termination_listener(sig=signal.SIGTERM, _registered=[]):
    if _registered:
        # already registered
        return
    _registered.append(True)

    def handle_signal(sig, stack):
        _logger.error("RED<<SIGNAL %s RECEIVED>>", sig)
        raise TerminationSignal(_signal=sig)

    if signal.getsignal(sig) in (signal.SIG_DFL, signal.SIG_IGN):  # check if signal is already trapped
        _logger.info("Listening to signal %s", sig)
        signal.signal(sig, handle_signal)
    else:
        raise SignalAlreadyBound(signal=sig)


def kill_subprocesses():
    from plumbum import local
    pid = os.getpid()
    return local.cmd.pkill['-HUP', '-P', pid].run(retcode=None)


def kill_this_process(graceful=False):
    from plumbum import local
    pid = os.getpid()
    if graceful:
        flag = '-HUP'
    else:
        flag = '-9'
    local.cmd.kill(flag, pid)


class Timebomb(object):

    def __init__(self, timeout, alert_interval=None, quiet=False):
        self.fuse = threading.Event()  # use this to cancel the timebomb
        self.timeout = timeout
        self.alert_interval = alert_interval
        self.quiet = quiet

    def __enter__(self):
        return self.start()

    def __exit__(self, *args):
        self.cancel()

    def start(self):
        from .concurrency import concurrent
        self.t = concurrent(self.wait_and_kill, daemon=True, threadname="Timebomb(%s)" % self.timeout)
        self.t.start()
        return self

    def cancel(self):
        self.fuse.set()

    @raise_in_main_thread()
    def wait_and_kill(self):
        timer = Timer(expiration=self.timeout)
        if not self.quiet:
            _logger.info("Timebomb set - this process will YELLOW<<self-destruct>> in RED<<%r>>...", timer.remain)
        while not timer.expired:
            if self.alert_interval:
                _logger.info("Time Elapsed: MAGENTA<<%r>>", timer.elapsed)
            log_level = logging.WARNING if timer.remain < 5 * MINUTE else logging.DEBUG
            _logger.log(log_level, "This process will YELLOW<<self-destruct>> in RED<<%r>>...", timer.remain)
            if self.fuse.wait(min(self.alert_interval or 60, timer.remain)):
                _logger.info("Timebomb cancelled")
                return
        _logger.warning("RED<< 💣 Timebomb Expired! 💣 >>")
        with _logger.indented("Killing children"):
            kill_subprocesses()
        with _logger.indented("Committing suicide"):
            kill_this_process(graceful=False)
            raise Exception("Timebomb Expired")


def set_timebomb(timeout, alert_interval=None):
    return Timebomb(timeout=timeout, alert_interval=alert_interval).start()


@wrapper_decorator
def shared_contextmanager(func):

    @locking_cache
    def inner(*args, **kwargs):

        class CtxManager():
            def __init__(self):
                self.count = 0
                self.func_cm = contextmanager(func)
                self._lock = threading.RLock()

            def __enter__(self):
                with self._lock:
                    if self.count == 0:
                        self.ctm = self.func_cm(*args, **kwargs)
                        self.obj = self.ctm.__enter__()
                    self.count += 1
                return self.obj

            def __exit__(self, *args):
                with self._lock:
                    self.count -= 1
                    if self.count > 0:
                        return
                    self.ctm.__exit__(*sys.exc_info())
                    del self.ctm
                    del self.obj

        return CtxManager()

    return inner


class TagAlongThread(object):

    def __init__(self, func, name, minimal_sleep=0):
        self._func = func
        self.minimal_sleep = minimal_sleep

        self._iteration_trigger = threading.Event()

        self._iterating = threading.Event()
        self._not_iterating = threading.Event()
        self._not_iterating.set()

        self._last_exception = None
        self._last_result = None

        self.__alive = True
        self._thread = threading.Thread(target=self._loop, daemon=True, name=name)
        self._thread.start()

    def _loop(self):
        while self.__alive:
            self._iteration_trigger.wait()
            self._iteration_trigger.clear()

            # Mark that we are now iterating
            self._not_iterating.clear()
            self._iterating.set()

            try:
                self._last_exception, self._last_result = None, self._func()
            except Exception as e:
                self._last_exception = e

            # Mark that we are no longer iterating
            self._iterating.clear()
            self._not_iterating.set()

            time.sleep(self.minimal_sleep)

        # Set all events so nothing will get blocked
        self._iteration_trigger.set()
        self._iterating.set()
        self._not_iterating.set()

    def __repr__(self):
        return 'TagAlongThread<%s>' % (self._thread.name,)

    def __call__(self):
        assert self.__alive, '%s is dead' % self
        # We can't use an iteration that's already started - maybe it's already at a too advanced stage?
        if self._iterating.is_set():
            self._not_iterating.wait()

        self._iteration_trigger.set()  # Signal that we want an iteration

        while not self._iterating.wait(1):  # Wait until an iteration starts
            # It is possible that we missed the loop and _iterating was already
            # cleared. If this is the case, _not_iterating will not be set -
            # and we can use it as a signal to stop waiting for iteration.
            if self._not_iterating.is_set():
                break
        else:
            self._not_iterating.wait()  # Wait until it finishes

        # To avoid races, copy last exception and result to local variables
        last_exception, last_result = self._last_exception, self._last_result
        if last_exception:
            raise last_exception
        else:
            return last_result

    def _kill(self, wait=True):
        self.__alive = False
        self._iteration_trigger.set()
        if wait:
            self._thread.join()


class SynchronizationCoordinatorWrongWait(TException):
    template = "Task is waiting on {this_file}:{this_line} instead of {others_file}:{others_line}"


class SynchronizationCoordinator(object):
    """
    Synchronization helper for functions that run concurrently::

        sync = SynchronizationCoordinator(5)

        def foo(a):
            sync.wait_for_everyone()
            sync.collect_and_call_once(a, lambda a_values: print(a))

        MultiObject(range(5)).call(foo)

    When MultiObject/concurrent_map/sync runs a function with a ``_sync=SYNC``
    argument, it will replace it with a proper SynchronizationCoordinator instance::

        def foo(a, _sync=SYNC):
            _sync.wait_for_everyone()
            _sync.collect_and_call_once(a, lambda a_values: print(a))

        MultiObject(range(5)).call(foo)
    """

    def __init__(self, num_participants):
        self.num_participants = num_participants
        self._reset_barrier()
        self._lock = threading.Lock()
        self._call_once_collected_param = []
        self._call_once_function = None
        self._call_once_result = None
        self._call_once_raised_exception = False

        self._wait_context = None

    def _reset_barrier(self):
        self.barrier = threading.Barrier(self.num_participants, action=self._post_barrier_action)

    def _post_barrier_action(self):
        if self.num_participants != self.barrier.parties:
            self._reset_barrier()
        self._wait_context = None

        if self._call_once_function:
            call_once_function, self._call_once_function = self._call_once_function, None
            collected_param, self._call_once_collected_param = self._call_once_collected_param, []

            try:
                self._call_once_result = call_once_function(collected_param)
                self._call_once_raised_exception = False
            except BaseException as e:
                self._call_once_result = e
                self._call_once_raised_exception = True

    def wait_for_everyone(self, timeout=HOUR):
        """
        Block until all threads that participate in the synchronization coordinator reach this point.

        Fail if one of the threads is waiting at a different point::

            def foo(a, _sync=SYNC):
                sleep(a)
                # Each thread will reach this point at a different time
                _sync.wait_for_everyone()
                # All threads will reach this point together

            MultiObject(range(5)).call(foo)
        """
        self._verify_waiting_on_same_line()
        self.barrier.wait(timeout=timeout)

    def abandon(self):
        """
        Stop participating in this synchronization coordinator.

        Note: when using with MultiObject/concurrent_map/asynchronous and _sync=SYNC, this
        is called automatically when a thread terminates on return or on exception.
        """
        with self._lock:
            self.num_participants -= 1
        self.barrier.wait()

    def collect_and_call_once(self, param, func, *, timeout=HOUR):
        """
        Call a function from one thread, with parameters collected from all threads::

            def foo(a, _sync=SYNC):
                result = _sync.collect_and_call_once(a, lambda a_values: set(a_values))
                assert result == {0, 1, 2, 3, 4}

            MultiObject(range(5)).call(foo)
        """
        self._verify_waiting_on_same_line()

        self._call_once_collected_param.append(param)
        self._call_once_function = func  # this will be set multiple times - but there is no race so that's OK

        self.barrier.wait(timeout=timeout)

        if self._call_once_raised_exception:
            raise self._call_once_result
        else:
            return self._call_once_result

    def _verify_waiting_on_same_line(self):
        frame = inspect.currentframe().f_back.f_back
        wait_context = (frame.f_code, frame.f_lineno, frame.f_lasti)

        existing_wait_context = self._wait_context
        if existing_wait_context is None:
            with self._lock:
                # Check again inside the lock, in case it was changed
                existing_wait_context = self._wait_context
                if existing_wait_context is None:
                    self._wait_context = wait_context
        if existing_wait_context is not None:  # could have changed inside the lock
            if wait_context != existing_wait_context:
                self.barrier.abort()
                raise SynchronizationCoordinatorWrongWait(
                    this_file=wait_context[0].co_filename,
                    this_line=wait_context[1],
                    others_file=existing_wait_context[0].co_filename,
                    others_line=existing_wait_context[1])

    def _abandon_when_done(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                result = func(*args, **kwargs)
                if hasattr(result, '__enter__') and hasattr(result, '__exit__'):
                    @contextmanager
                    def wrapper_cm():
                        try:
                            with result as yielded_value:
                                self.wait_for_everyone()  # see https://github.com/weka-io/easypy/issues/150
                                yield yielded_value
                        finally:
                            self.abandon()
                    return wrapper_cm()
                else:
                    self.abandon()
                    return result
            except:
                self.abandon()
                raise
        return wrapper


class SYNC(SynchronizationCoordinator):
    """Mimic ``SynchronizationCoordinator`` for running in single thread."""

    def __init__(self):
        pass

    def wait_for_everyone(self):
        pass
    wait_for_everyone.__doc__ = SynchronizationCoordinator.wait_for_everyone.__doc__

    def abandon(self):
        pass
    abandon.__doc__ = SynchronizationCoordinator.abandon.__doc__

    def collect_and_call_once(self, param, func):
        return func([param])
    collect_and_call_once.__doc__ = SynchronizationCoordinator.collect_and_call_once.__doc__


SYNC = SYNC()


def _get_my_ident():
    return Hex(threading.current_thread().ident)


class LoggedRLock():
    """
    Like ``RLock``, but more logging friendly.

    :param name: give it a name, so it's identifiable in the logs
    :param log_interval: the interval between log messages
    :param lease_expiration: throw an exception if the lock is held for more than this duration
    """

    # we could inherit from this and support other types, but that'll require changes in the repr
    LockType = threading.RLock

    __slots__ = ("_lock", "_name", "_lease_expiration", "_lease_timer", "_log_interval", "_get_data")
    _RE_OWNER = re.compile(r".*owner=(\d+) count=(\d+).*")
    _MIN_TIME_FOR_LOGGING = 10

    def __init__(self, name=None, log_interval=15, lease_expiration=NEVER):
        self._lock = self.__class__.LockType()
        self._name = name or '{}-{:X}'.format(self.LockType.__name__, id(self))
        self._lease_expiration = lease_expiration
        self._lease_timer = None
        self._log_interval = log_interval

        # we want to support both the gevent and builtin lock
        try:
            self._lock._owner
        except AttributeError:
            def _get_data():
                return tuple(map(int, self._RE_OWNER.match(repr(self._lock)).groups()))
        else:
            def _get_data():
                return self._lock._owner, self._lock._count
        self._get_data = _get_data

    def __repr__(self):
        owner, count = self._get_data()
        try:
            owner = threading._active[owner].name
        except KeyError:
            pass
        if owner:
            return "<{}, owned by <{}>x{} for {}>".format(self._name, owner, count, self._lease_timer.elapsed)
        else:
            return "<{}, unowned>".format(self._name)

    def _acquired(self, lease_expiration, should_log=False):
        # we don't want to replace the lease timer, so not to effectively extend the original lease
        self._lease_timer = self._lease_timer or Timer(expiration=lease_expiration or self._lease_expiration)
        if should_log:
            _logger.debug("%s - acquired", self)

    def acquire(self, blocking=True, timeout=-1, lease_expiration=None):
        if not blocking:
            ret = self._lock.acquire(blocking=False)
            # touch it once, so we don't hit a race since it occurs outside of the lock acquisition
            lease_timer = self._lease_timer
            if ret:
                self._acquired(lease_expiration)
            elif lease_timer and lease_timer.expired:
                raise LockLeaseExpired(lock=self)
            return ret

        # this timer implements the 'timeout' parameter
        acquisition_timer = Timer(expiration=NEVER if timeout < 0 else timeout)
        while not acquisition_timer.expired:

            # the timeout on actually acquiring this lock is the minimum of:
            # 1. the time remaining on the acquisition timer, set by the 'timeout' param
            # 2. the logging interval - the minimal frequency for logging while the lock is awaited
            # 3. the time remaining on the lease timer, which would raise if expired
            timeout = min(acquisition_timer.remain, self._log_interval)
            # touch it once, so we don't hit a race since it occurs outside of the lock acquisition
            lease_timer = self._lease_timer
            if lease_timer:
                timeout = min(lease_timer.remain, timeout)

            if self._lock.acquire(blocking=True, timeout=timeout):
                self._acquired(lease_expiration, should_log=acquisition_timer.elapsed > self._MIN_TIME_FOR_LOGGING)
                return True

            # touch it once, so we don't hit a race since it occurs outside of the lock acquisition
            lease_timer = self._lease_timer
            if lease_timer and lease_timer.expired:
                raise LockLeaseExpired(lock=self)

            _logger.debug("%s - waiting...", self)

    def release(self, *args):
        _, count = self._get_data()
        if count == 1:
            # we're last: clear the timer before releasing the lock!
            if self._lease_timer.elapsed > self._MIN_TIME_FOR_LOGGING:
                _logger.debug("%s - releasing...", self)
            self._lease_timer = None
        self._lock.release()

    __exit__ = release
    __enter__ = acquire


class RWLock(object):
    """
    Read-Write Lock: allows locking exclusively and non-exclusively::

        rwl = RWLock()

        with rwl:
            # other can acquire this lock, but not exclusively

        with rwl.exclusive():
            # no one can acquire this lock - we are alone here

    """

    def __init__(self, name=None):
        self.lock = threading.RLock()
        self.cond = threading.Condition(self.lock)
        self.owners = Counter()
        self.name = name or '{}-{:X}'.format(self.__class__.__name__, id(self.lock))
        self._lease_timer = None

    def __repr__(self):
        owners = ", ".join(map(str, sorted(self.owners.keys())))
        lease_timer = self._lease_timer  # touch once to avoid races
        if lease_timer:
            mode = "exclusively ({})".format(lease_timer.elapsed)
        else:
            mode = "non-exclusively"
        return "<{}, owned by <{}> {}>".format(self.name, owners, mode)

    @property
    def owner_count(self):
        return sum(self.owners.values())

    def __call__(self):
        return self

    def __enter__(self):
        while not self.cond.acquire(timeout=15):
            _logger.debug("%s - waiting...", self)

        try:
            self.owners[_get_my_ident()] += 1
            _verbose_logger.debug("%s - acquired (non-exclusively)", self)
            return self
        finally:
            self.cond.release()

    def __exit__(self, *args):
        while not self.cond.acquire(timeout=15):
            _logger.debug("%s - waiting...", self)

        try:
            my_ident = _get_my_ident()
            self.owners[my_ident] -= 1
            if not self.owners[my_ident]:
                self.owners.pop(my_ident)  # don't inflate the soft lock keys with threads that does not own it
            self.cond.notify()
            _verbose_logger.debug("%s - released (non-exclusive)", self)
        finally:
            self.cond.release()

    @contextmanager
    def exclusive(self, need_to_wait_message=None):
        while not self.cond.acquire(timeout=15):
            _logger.debug("%s - waiting...", self)

        # wait until this thread is the sole owner of this lock
        while not self.cond.wait_for(lambda: self.owner_count == self.owners[_get_my_ident()], timeout=15):
            _check_exiting()
            if need_to_wait_message:
                _logger.info(need_to_wait_message)
                need_to_wait_message = None  # only print it once
            _logger.debug("%s - waiting (for exclusivity)...", self)
        my_ident = _get_my_ident()
        self.owners[my_ident] += 1
        self._lease_timer = Timer()
        _verbose_logger.debug("%s - acquired (exclusively)", self)
        try:
            yield
        finally:
            _verbose_logger.debug('%s - releasing (exclusive)', self)
            self._lease_timer = None
            self.owners[my_ident] -= 1
            if not self.owners[my_ident]:
                self.owners.pop(my_ident)  # don't inflate the soft lock keys with threads that does not own it
            self.cond.notify()
            self.cond.release()


SoftLock = RWLock


@parametrizeable_decorator
def skip_if_locked(func=None, lock=None, default=None):
    if not lock:
        lock = threading.RLock()

    def inner(*args, **kwargs):
        if not lock.acquire(blocking=False):
            _logger.debug("lock acquired - skipped %s", func)
            return default
        try:
            return func(*args, **kwargs)
        finally:
            lock.release()
    return inner


@parametrizeable_decorator
def with_my_lock(method, lock_attribute="_lock"):

    def inner(self, *args, **kwargs):
        ctx = getattr(self, lock_attribute)
        if callable(ctx):
            ctx = ctx()
        with ctx:
            return method(self, *args, **kwargs)

    return inner


@parametrizeable_decorator
def synchronized(func, lock=None):
    if not lock:
        lock = threading.RLock()

    def inner(*args, **kwargs):
        with lock:
            return func(*args, **kwargs)

    return inner


class SynchronizedSingleton(type):
    _instances = {}

    @synchronized
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)

        return cls._instances[cls]

    @synchronized
    def get_instance(cls):
        return cls._instances.get(cls)


class LoggedCondition():
    """
    Like Condition, but easier to use and more logging friendly

    :param name: give it a name, so it's identifiable in the logs
    :param log_interval: the interval between log messages

    Unlike threading.condition, .acquire() and .release() are not needed here.
    Just use .wait_for() to wait for a predicate and perform the
    condition-changing statements inside a .notifying_all() context::

        some_flag = False
        cond = Condition('some flag cond')

        # Wait for condition:
        cond.wait_for(lambda: some_flag, 'some flag become True')

        # Trigger the condition:
        with cond.notifying_all('Setting some flag to true'):
            some_flag = True
    """

    ConditionType = threading.Condition

    __slots__ = ("_cond", "_name", "_log_interval")

    def __init__(self, name=None, log_interval=15):
        self._cond = self.__class__.ConditionType()
        self._name = name or '{}-{:X}'.format(self.ConditionType.__name__, id(self))
        self._log_interval = log_interval

    def __repr__(self):
        return '<{}>'.format(self._name)

    @contextmanager
    def _acquired_for(self, msg, *args):
        while not self._cond.acquire(timeout=self._log_interval):
            _logger.debug('%s - waiting to be acquired for ' + msg, self, *args)
        try:
            yield
        finally:
            self._cond.release()

    @contextmanager
    def notifying_all(self, msg, *args):
        """
        Acquire the condition lock for the context, and notify all waiters afterward.

        :param msg: Message to print to the DEBUG log after performing the command
        :param args: Format arguments for msg

        Users should run the command that triggers the conditions inside this context manager.
        """
        with self._acquired_for('performing a %s notifying all waiters' % msg, *args):
            yield
            _logger.debug('%s - performed: ' + msg, self, *args)
            self._cond.notifyAll()

    @contextmanager
    def __wait_for_impl(self, pred, msg, *args, timeout=None):
        timer = Timer(expiration=timeout)

        def timeout_for_condition():
            remain = timer.remain
            if remain:
                return min(remain, self._log_interval)
            else:
                return self._log_interval

        with self._acquired_for('checking ' + msg, *args):
            while not self._cond.wait_for(pred, timeout=timeout_for_condition()):
                if timer.expired:
                    # NOTE: without a timeout we will never get here
                    if pred():  # Try one last time, to make sure the last check was not (possibly too long) before the timeout
                        return
                    raise TimeoutException('{condition} timed out after {duration} waiting for {msg}',
                                           condition=self, msg=msg % args, duration=timer.elapsed)
                _logger.debug('%s - waiting for ' + msg, self, *args)
            yield

    def wait_for(self, pred, msg, *args, timeout=None):
        """
        Wait for a predicate. Only check it when notified.

        :param msg: Message to print to the DEBUG log while waiting for the predicate
        :param args: Format arguments for msg
        :param timeout: Maximal time to wait
        """
        with self.__wait_for_impl(pred, msg, *args, timeout=timeout):
            pass

    @contextmanager
    def waited_for(self, pred, msg, *args, timeout=None):
        """
        Wait for a predicate, keep the condition lock for the context, and notify all other waiters afterward.

        :param msg: Message to print to the DEBUG log while waiting for the predicate
        :param args: Format arguments for msg
        :param timeout: Maximal time to wait

        The code inside the context should be used for altering state other waiters are waiting for.
        """
        with self.__wait_for_impl(pred, msg, *args, timeout=timeout):
            yield
            self._cond.notifyAll()

    @property
    def lock(self):
        """
        Use the underlying lock without notifying the waiters.
        """

        return self._cond._lock


# cache result only when predicate succeeds
class CachingPredicate():
    def __init__(self, pred):
        self.pred = pred

    def __call__(self, *args, **kwargs):
        try:
            return self.result
        except AttributeError:
            pass
        ret = self.pred(*args, **kwargs)
        if ret in (False, None):
            return ret
        self.result = ret
        return self.result


def make_multipred(preds):
    preds = list(map(CachingPredicate, preds))

    def pred(*args, **kwargs):
        results = [pred(*args, **kwargs) for pred in preds]
        if all(results):
            return results
    return pred


def iter_wait(
        timeout, pred=None, sleep=0.5, message=None,
        progressbar=True, throw=True, allow_interruption=False, caption=None,
        log_interval=10 * MINUTE, log_level=logging.DEBUG):

    # Calling wait() with a predicate and no message is very not informative
    # (throw=False or message=False disables this behavior)
    if message is False:
        message = None
    elif throw and pred and not message:
        raise Exception(
            "Function iter_wait()'s parameter `message` is required if "
            "`pred` is passed",
        )

    if timeout is None:
        msg = "Waiting indefinitely%s"
    else:
        msg = "Waiting%%s up to %s" % time_duration(timeout)

    if message is None:
        if caption:
            message = "Waiting %s timed out after {duration:.1f} seconds" % (caption,)
        elif pred:
            message = "Waiting on predicate (%s) timed out after {duration:.1f} seconds" % (pred,)
        else:
            message = "Timed out after {duration:.1f} seconds"

    if pred:
        pred_decorator = kwargs_resilient(negligible=['is_final_attempt'])
        if hasattr(pred, "__iter__"):
            pred = make_multipred(map(pred_decorator, pred))
        else:
            pred = pred_decorator(pred)
        if not caption:
            caption = "on predicate (%s)" % pred
    else:
        pred = lambda **kwargs: False
        throw = False

    if caption:
        msg %= " %s" % (caption,)
    else:
        msg %= ""

    if isinstance(sleep, tuple):
        data = list(sleep)  # can't use nonlocal since this module is indirectly used in python2

        def sleep():
            cur, mx = data
            try:
                return cur
            finally:
                data[0] = min(mx, cur * 1.5)

    if not IS_A_TTY:
        # can't interrupt
        allow_interruption = False
        progressbar = False

    if progressbar and threading.current_thread() is not threading.main_thread():
        # prevent clutter
        progressbar = False

    if allow_interruption:
        msg += " (hit <ESC> to continue)"

    l_timer = Timer(expiration=timeout)
    log_timer = Timer(expiration=log_interval)

    with ExitStack() as stack:
        if progressbar:
            from .logging import PROGRESS_BAR
            pr = stack.enter_context(PROGRESS_BAR())
            pr.set_message(msg)

        while True:
            s_timer = Timer()
            expired = l_timer.expired
            last_exc = None
            try:
                ret = pred(is_final_attempt=bool(expired))
            except PredicateNotSatisfied as _exc:
                if getattr(_exc, "duration", 0):
                    # this exception was raised by a nested 'wait' call - don't swallow it!
                    raise
                if log_timer.expired:
                    log_timer.reset()
                    _logger.log(log_level, 'Still waiting after %r: %s', l_timer.elapsed, _exc.message)
                last_exc = _exc
                ret = None
            else:
                if ret not in (None, False):
                    yield ret
                    return
            if expired:
                duration = l_timer.stop()
                start_time = l_timer.start_time
                if throw:
                    if last_exc:
                        last_exc.add_params(duration=duration, start_time=start_time)
                        raise last_exc
                    if callable(message):
                        message = message()
                    raise TimeoutException(message, duration=duration, start_time=start_time)
                yield None
                return
            yield l_timer.remain
            sleep_for = sleep() if callable(sleep) else sleep
            if allow_interruption:
                from termenu.keyboard import keyboard_listener
                timer = Timer(expiration=sleep_for - s_timer.elapsed)
                for key in keyboard_listener(heartbeat=0.25):
                    if key == "esc":
                        yield None
                        return
                    if key == "enter":
                        break
                    if timer.expired:
                        break
            else:
                s_timeout = max(0, sleep_for - s_timer.elapsed)
                if l_timer.expiration:
                    s_timeout = min(l_timer.remain, s_timeout)
                time.sleep(s_timeout)


@wraps(iter_wait)
def wait(*args, **kwargs):
    """
    Wait until ``pred`` returns a useful value (see below), or until ``timeout`` passes.

    :param timeout: how long to wait for ``pred`` to get satisfied. if ``None`` waits
        indefinitely.
    :param pred: callable that checks the condition to wait upon.
        It can return ``None`` or ``False`` to indicate that the predicate has not been satisfied.
        Any other value will end the wait and that value will be returned from the wait function.
        The predicate can also raise a subclass of PredicateNotSatisfied. The exception will be raised
        if the timeout expires, instead of a TimeoutException.
        ``pred`` can be a list of predicates, and ``wait`` will wait for all of them to be satisfied.
        Note that once a predicate is satisfied, it will not be called again.
        If no ``pred`` is provided, ``wait`` behaves like ``sleep``.
    :param sleep: the number of seconds to sleep between calls to ``pred``.
        it can be a callable, or a ``(first, max)`` tuple, in which case the sleep duration will grow
        exponentially from ``first`` up to ``max``.
    :param message: message to use for a TimeoutException. can be a callable. To encourage the use of
        informative TimeoutException messages, the user must provide a value here. If a PredicateNotSatisfied
        is used in the predicate, pass ``False``.
    :param progressbar: if True, show an automatic progress bar while waiting.
    :param caption: message to show in progress bar, and in TimeoutException (if ``message``
        not given).
    :param throw: if True, throw an exception if ``timeout`` expires.
        if ``pred`` not given, this is always False.
    :param allow_interruption: if True, the user can end the wait prematurely by hitting ESC.
    :param log_interval: interval for printing thrown ``PredicateNotSatisfied``s to the log.
        Set to ``None`` to disable this logging. If the predicate returns ``False`` instead
        of throwing this argument will be ignored. Defaults to 10 minutes.
    :param log_level: the log level for printing the thrown ``PredicateNotSatisfied`` with
        ``log_interval``. Defaults to ``logging.DEBUG``.
    """
    for ret in iter_wait(*args, **kwargs):
        pass
    return ret


def wait_progress(*args, **kwargs):
    for _ in iter_wait_progress(*args, **kwargs):
        pass


def iter_wait_progress(state_getter, advance_timeout, total_timeout=float("inf"), state_threshold=0, sleep=0.5, throw=True,
                       allow_regression=True, advancer_name=None, progressbar=True):

    ADVANCE_TIMEOUT_MESSAGE = "did not advance for {duration: .1f} seconds"
    TOTAL_TIMEOUT_MESSAGE = "advanced but failed to finish in {duration: .1f} seconds"

    state = state_getter()  # state_getter should return a number, represent current state

    progress = Bunch(state=state, finished=False, changed=False)
    progress.total_timer = Timer(expiration=total_timeout)
    progress.advance_timer = Timer(expiration=advance_timeout)

    def did_advance():
        current_state = state_getter()
        progress.advanced = progress.state > current_state
        progress.changed = progress.state != current_state
        if progress.advanced or allow_regression:
            progress.state = current_state
        return progress.advanced

    # We want to make sure each iteration sleeps at least once,
    # since the internal 'wait' could return immediately without sleeping at all,
    # and if the external while loop isn't done we could be iterating too much
    min_sleep = None

    while progress.state > state_threshold:
        progress.timeout, message = min(
            (progress.total_timer.remain, TOTAL_TIMEOUT_MESSAGE),
            (progress.advance_timer.remain, ADVANCE_TIMEOUT_MESSAGE))
        if advancer_name:
            message = advancer_name + ' ' + message

        if min_sleep:
            wait(min_sleep.remain)
        min_sleep = Timer(expiration=sleep)

        result = wait(progress.timeout, pred=did_advance, sleep=sleep, message=message, throw=throw, progressbar=progressbar)
        if not result:  # if wait times out without throwing
            return

        if progress.total_timer.expired:
            raise TimeoutException(message, duration=progress.total_timer.duration)

        progress.advance_timer.reset()
        yield progress

    progress.finished = True
    yield progress  # indicate success


from .timing import Timer  # noqa; avoid import cycle



# encoding: utf-8

"""
This module helps you run things concurrently.
Most useful are the ``concurrent`` context manager and the ``MultiObject`` class.
The features in this module are integrated with the ``logging`` module, to provide
thread-context to log messages. It also has support for integration ``gevent``.

``MultiObject``

Here's how you would, for example, concurrently send a message to a bunch of servers::

    responses = MultiObject(servers).send('Hello')

The above replaces the following threading boiler-plate code::

    from threading import ThreadPoolExecutor

    with ThreadPoolExecutor(max_workers=3) as executor:
        futures = [
            executor.submit(server.send, args=['Hello'])
            for server in servers]

    responses = [future.result() for future in futures]```


``concurrent``

A high-level thread controller.
As a context manager, it runs a function in a thread, and joins it upon exiting the context

    with concurrent(requests.get, "api.myserver.com/data") as future:
        my_data = open("local_data").read()

    remote_data = future.result()
    process_it(my_data, remote_data)

It can also be used to run something repeatedly in the background:

    concurrent(send_heartbeat, loop=True, sleep=5).start()  # until the process exits

    with concurrent(print_status, loop=True, sleep=5):
        some_long_task()


Environment variables

    EASYPY_DISABLE_CONCURRENCY (yes|no|true|false|1|0)
    EASYPY_MAX_THREAD_POOL_SIZE

Important notes

    * Exception.timestamp

    Exceptions raised in functions that use the features here may get a ``timestamp`` attribute that
    records the precise time those exceptions were raised. This is useful when there's a lag between
    when the exception was raised within a thread, and when that exception was finally propagated to
    the calling thread.

"""


from concurrent.futures import ThreadPoolExecutor, CancelledError, as_completed, Future, wait as futures_wait
from concurrent.futures import TimeoutError as FutureTimeoutError

from collections import defaultdict
from contextlib import contextmanager, ExitStack
from functools import partial, wraps
from importlib import import_module
from itertools import chain
from traceback import format_tb
import inspect
import logging
import threading
import time
import os
from collections import namedtuple
from datetime import datetime

import easypy._multithreading_init  # noqa; make it initialize the threads tree
from easypy.exceptions import PException, apply_timestamp
from easypy.gevent import is_module_patched, non_gevent_sleep, defer_to_thread
from easypy.humanize import IndentableTextBuffer, time_duration, compact
from easypy.humanize import format_thread_stack, yesno_to_bool
from easypy.threadtree import iter_thread_frames
from easypy.timing import Timer
from easypy.units import MINUTE, HOUR
from easypy.colors import colorize, uncolored
from easypy.sync import SynchronizationCoordinator, ProcessExiting, raise_in_main_thread


MAX_THREAD_POOL_SIZE = int(os.environ.get('EASYPY_MAX_THREAD_POOL_SIZE', 50))
DISABLE_CONCURRENCY = yesno_to_bool(os.getenv("EASYPY_DISABLE_CONCURRENCY", "no"))

this_module = import_module(__name__)
THREADING_MODULE_PATHS = [threading.__file__]

IS_GEVENT = False

if is_module_patched("threading"):
    IS_GEVENT = True
    import gevent
    MAX_THREAD_POOL_SIZE *= 100  # these are not threads anymore, but greenlets. so we allow a lot of them
    THREADING_MODULE_PATHS.append(gevent.__path__[0])


try:
    from traceback import _extract_stack_iter
except ImportError:
    from traceback import walk_stack

    def _extract_stack_iter(frame):
        for f, lineno in walk_stack(frame):
            co = f.f_code
            filename = co.co_filename
            name = co.co_name
            yield filename, lineno, name


_logger = logging.getLogger(__name__)


def disable():
    """
    Force MultiObject and concurrent calls to run synchronuously in the current thread.
    For debugging purposes.
    """
    global DISABLE_CONCURRENCY
    DISABLE_CONCURRENCY = True
    logging.info("Concurrency disabled")


def enable():
    """
    Re-enable concurrency, after disabling it
    """
    global DISABLE_CONCURRENCY
    DISABLE_CONCURRENCY = False
    logging.info("Concurrency enabled")


def _find_interesting_frame(f):
    """
    Find the next frame in the stack that isn't threading-related, to get to the actual caller.
    """
    default = next(_extract_stack_iter(f))
    non_threading = (
        p for p in _extract_stack_iter(f)
        if all(not p[0].startswith(pth) for pth in THREADING_MODULE_PATHS))
    return next(non_threading, default)


class MultiExceptionMeta(type):
    """
    This metaclass helps generate MultiException subtypes so that it's easier
    to catch a MultiException with a specific common type, for example::

        try:
            MultiObject(servers).connect()
        except MultiException[ConnectionError]:
            pass

    See ``MultiException`` for more information. 
    """

    _SUBTYPES = {}
    _SUBTYPES_LOCK = threading.RLock()

    def __getitem__(cls, exception_type):
        if exception_type is BaseException:
            return MultiException

        assert isinstance(exception_type, type), "Must use an Exception type"
        assert issubclass(exception_type, BaseException), "Must inherit for BaseException"

        try:
            return cls._SUBTYPES[exception_type]
        except KeyError:
            with cls._SUBTYPES_LOCK:
                if exception_type in cls._SUBTYPES:
                    return cls._SUBTYPES[exception_type]
                bases = tuple(cls[base] for base in exception_type.__bases__ if base and issubclass(base, BaseException))
                subtype = type("MultiException[%s]" % exception_type.__qualname__, bases, dict(COMMON_TYPE=exception_type))
                cls._SUBTYPES[exception_type] = subtype
                return subtype

    __iter__ = None

    def __call__(cls, exceptions, futures):
        common_type = concestor(*map(type, filter(None, exceptions)))
        subtype = cls[common_type]
        return type.__call__(subtype, exceptions, futures)


PickledFuture = namedtuple("PickledFuture", "ctx, funcname")


class MultiException(PException, metaclass=MultiExceptionMeta):
    """
    A ``MultiException`` subtype is raised when a ``MultiObject`` call fails in one or more of its threads.
    The exception contains the following members:

    :param actual: a MultiObject of all exceptions raised in the ``MultiObject`` call
    :param count: the number of threads that raised an exception
    :param invocations_count: the total number of calls the ``MultiObject`` made (the size of the ``MultiObject``)
    :param common_type: the closest common ancestor (base-class) of all the exceptions
    :param one: a sample exception (the first)
    :param futures: a MultiObject of futures (:concurrent.futures.Future:) that were created in the ``MultiObject`` call
    :param exceptions: a sparse list of exceptions corresponding to the MultiObject threads
    :param complete: ``True`` if all threads failed on exception
    """

    template = "{0.common_type.__qualname__} raised from concurrent invocation (x{0.count}/{0.invocations_count})"

    def __reduce__(self):
        return (MultiException, (self.exceptions, [PickledFuture(ctx=f.ctx, funcname=f.funcname) for f in self.futures]))

    def __init__(self, exceptions, futures):
        # we want to keep futures in parallel with exceptions,
        # so some exceptions could be None
        assert len(futures) == len(exceptions)
        self.actual = MultiObject(filter(None, exceptions))
        self.count = len(self.actual)
        self.invocations_count = len(futures)
        self.common_type = self.COMMON_TYPE
        self.one = self.actual.T[0] if self.actual else None
        self.futures = MultiObject(futures)
        self.exceptions = exceptions
        self.complete = self.count == self.invocations_count
        if self.complete and hasattr(self.common_type, 'exit_with_code'):
            self.exit_with_code = self.common_type.exit_with_code
        super().__init__(self.template, self)

    def __repr__(self):
        return "{0.__class__.__name__}(x{0.count}/{0.invocations_count})".format(self)

    def __str__(self):
        return self.render(color=False)

    def walk(self, skip_multi_exceptions=True):
        if not skip_multi_exceptions:
            yield self
        for exc in self.actual:
            if isinstance(exc, MultiException):
                yield from exc.walk(skip_multi_exceptions=skip_multi_exceptions)
            else:
                yield exc

    def render(self, *, width=80, color=True, **kw):
        buff = self._get_buffer(color=color, **kw)
        text = buff.render(width=width, edges=not color)
        return colorize("\n" + text)

    def _get_buffer(self, **kw):
        if kw.get("color", True):
            normalize_color = lambda x: x
        else:
            normalize_color = uncolored

        def _format_context(context):
            if not isinstance(context, dict):
                return repr(context)  # when it comes from rpyc
            context = context.copy()
            context.pop("indentation", None)
            breadcrumbs = ";".join(context.pop('context', []))
            return ", ".join(filter(None, chain((breadcrumbs,), ("%s=%s" % p for p in sorted(context.items())))))

        buff = IndentableTextBuffer("{0.__class__.__qualname__}", self)
        if self.message:
            buff.write(normalize_color("WHITE<<%s>>" % self.message))

        traceback_fmt = normalize_color("DARK_GRAY<<{}>>")

        # workaround incompatibilty with rpyc, which serializes .actual into an str
        # instead of a list of exceptions. This makes the string flatten into a long
        # and incomprehensible text buffer.
        if hasattr(self, "_remote_tb"):
            with buff.indent("Remote Traceback:"):
                buff.write(self._remote_tb)
            return buff

        def add_details(exc):
            if kw.get("timestamp", True) and getattr(exc, "timestamp", None):
                ts = datetime.fromtimestamp(exc.timestamp).isoformat()
                buff.write(normalize_color("MAGENTA<<Timestamp: %s>>" % ts))
            if kw.get("context", True) and getattr(exc, "context", None):
                buff.write("Context: %s" % _format_context(exc.context))

        add_details(self)

        for exc in self.actual:
            with buff.indent("{.__class__.__qualname__}", exc):
                if isinstance(exc, MultiException):
                    buff.extend(exc._get_buffer(**kw))
                elif callable(getattr(exc, "render", None)):
                    buff.write(exc.render(**kw))
                else:
                    buff.write("{}", exc)
                    add_details(exc)
                if hasattr(exc, "__traceback__"):
                    show_traceback = getattr(exc, 'traceback', None)
                    if show_traceback is not False:
                        buff.write("Traceback:")
                        for line in format_tb(exc.__traceback__):
                            buff.write(traceback_fmt, line.rstrip())

        return buff


def _submit_execution(executor, func, args, kwargs, ctx, funcname=None):
    """
    This helper takes care of submitting a function for asynchronous execution, while wrapping and storing
    useful information for tracing it in logs (for example, by ``Futures.dump_stacks``)
    """
    future = executor.submit(_run_with_exception_logging, func, args, kwargs, ctx)
    future.ctx = ctx
    future.funcname = funcname or _get_func_name(func)
    return future


class Futures(list):
    """
    A collection of ``Future`` objects.
    """

    def done(self):
        """
        Return ``True`` if all futures are done
        """
        return all(f.done() for f in self)

    def cancelled(self):
        """
        Return ``True`` if all futures are cancelled
        """
        return all(f.cancelled() for f in self)

    def running(self):
        """
        Return ``True`` if all futures are running
        """
        return all(f.running() for f in self)

    def wait(self, timeout=None):
        """
        Wait for all Futures to complete
        """
        return futures_wait(self, timeout=timeout)

    def result(self, timeout=None):
        """
        Wait and return the results from all futures as an ordered list.
        Raises a ``MultiException`` if one or more exceptions are raised.
        """
        me = self.exception(timeout=timeout)
        if me:
            if isinstance(me, MultiException[ProcessExiting]):
                # we want these aborted MultiObject threads to consolidate this exception
                raise ProcessExiting()
            raise me
        return [f.result() for f in self]

    def exception(self, timeout=None):
        """
        Wait and return a ``MultiException`` if there any exceptions, otherwise returns ``None``.
        """
        exceptions = [f.exception(timeout=timeout) for f in self]
        if any(exceptions):
            return MultiException(exceptions=exceptions, futures=self)

    def cancel(self):
        """
        Cancel all futures.
        """
        cancelled = [f.cancel() for f in self]  # list-comp, to ensure we call cancel on all futures
        return all(cancelled)

    def as_completed(self, timeout=None):
        """
        Returns an iterator yielding the futures in order of completion.
        Wraps `concurrent.futures.as_completed`.
        """
        return as_completed(self, timeout=timeout)

    @classmethod
    @contextmanager
    def execution(cls, workers=None, ctx={}):
        """
        A context-manager for scheduling asynchronous executions and waiting on them as upon exiting the context::

            With Futures.execution() as futures:
                for task in tasks:
                    futures.submit(task)

            results = futures.results()
        """
        if workers is None:
            workers = MAX_THREAD_POOL_SIZE

        class PooledFutures(cls):

            killed = False

            def submit(self, func, *args, log_ctx={}, **kwargs):
                "Submit a new asynchronous task to this executor"

                _ctx = dict(ctx, **log_ctx)
                future = executor.submit(_run_with_exception_logging, func, args, kwargs, _ctx)
                future.ctx = _ctx
                future.funcname = _get_func_name(func)
                self.append(future)
                return future

            def kill(self):
                "Kill the executor and discard any running tasks"
                self.cancel()
                self.shutdown(wait=False)
                while executor._threads:
                    thread = executor._threads.pop()
                    if getattr(thread, "_greenlet", None):
                        thread._greenlet.kill()
                self.killed = True

            def shutdown(self, *args, **kwargs):
                executor.shutdown(*args, **kwargs)

        with ThreadPoolExecutor(workers) as executor:

            futures = PooledFutures()

            try:
                yield futures
            except:  # noqa
                _logger.debug("shutting down ThreadPoolExecutor due to exception")
                futures.kill()
                raise
            else:
                if not futures.killed:
                    # force exceptions to bubble up
                    futures.result()
            finally:
                # break the cycle so that the GC doesn't clean up the executor under a lock (https://bugs.python.org/issue21009)
                futures.kill = futures.shutdown = futures.submit = None
                futures = None

    executor = execution  # for backwards compatibility

    @classmethod
    def dump_stacks(cls, futures, verbose=False):
        """
        Logs the stack frame for each of the given futures.
        The Future objects must have been submitted with ``_submit_execution`` so that they contain
        the necessary information.
        """
        frames = dict(iter_thread_frames())
        for i, future in enumerate(futures, 1):
            thread_ident = future.ctx['thread_ident']
            try:
                frame = frames[thread_ident]
            except KeyError:
                assert False, frames
                frame = None  # this might happen in race-conditions with a new thread starting
            if not verbose or not frame:
                if frame:
                    frame_line = _find_interesting_frame(frame)[:3]
                    location = " - %s:%s, in %s(..)" % tuple(frame_line)
                else:
                    location = "..."
                _logger.info("%3s - %s (DARK_YELLOW<<%s>>)%s",
                             i, future.funcname, cls._get_context(future), location)
                continue

            with _logger.indented("%3s - %s (%s)", i, future.funcname, cls._get_context(future), footer=False):
                lines = format_thread_stack(frame, skip_modules=[this_module]).splitlines()
                for line in lines:
                    _logger.info(line.strip())

    @classmethod
    def _get_context(cls, future: Future):
        """
        Get interesting context information about this future object (as long as it was submitted by _submit_execution)
        """
        def compacted(s):
            return compact(str(s).split("\n", 1)[0], 20, "....", 5).strip()
        ctx = dict(future.ctx)
        context = []
        threadname = ctx.pop("threadname", None)
        thread_ident = ctx.pop("thread_ident", None)
        context.append(threadname or thread_ident)
        context.append(ctx.pop("context", None))
        context.extend("%s=%s" % (k, compacted(v)) for k, v in sorted(ctx.items()))
        return ";".join(filter(None, context))

    def logged_wait(self, timeout=None, initial_log_interval=2 * MINUTE):
        """
        Wait for all futures to complete, logging their status along the way.
        Logging will occur at an every-increasing log interval, beginning with ``initial_log_interval``,
        and increasing 5-fold (x5) every 5 iterations.
        """
        log_interval = initial_log_interval
        global_timer = Timer(expiration=timeout)
        iteration = 0

        while not global_timer.expired:
            completed, pending = self.wait(log_interval)
            if not pending:
                break

            iteration += 1
            if iteration % 5 == 0:
                log_interval *= 5
            with _logger.indented("(Waiting for %s on %s/%s tasks...)",
                                  time_duration(global_timer.elapsed),
                                  len(pending), sum(map(len, (completed, pending))),
                                  level=logging.WARNING, footer=False):
                self.dump_stacks(pending, verbose=global_timer.elapsed >= HOUR)


def _run_with_exception_logging(func, args, kwargs, ctx):
    """
    Use as a wrapper for functions that run asynchronously, setting up a logging context and
    recording the thread in-which they are running, so that we can later log their progress
    and identify the source of exceptions they raise. In addition, it stamps any exception
    raised from the function with the current time.
    """
    thread = threading.current_thread()
    ctx.update(threadname=thread.name, thread_ident=thread.ident)
    with _logger.context(**ctx):
        try:
            return func(*args, **kwargs)
        except StopIteration:
            # no need to log this
            raise
        except ProcessExiting as exc:
            _logger.debug(exc)
            raise
        except Exception as exc:
            _logger.silent_exception(
                "Exception (%s) in thread running %s (traceback in debug logs)",
                exc.__class__.__qualname__, func)
            try:
                exc.timestamp = time.time()
            except:  # noqa - sometimes exception objects are immutable
                pass
            raise


def _to_args_list(params):
    "Helper for normalizing a list of parameters to be mapped on a function"
    # We use type(args) == tuple because isinstance will return True for namedtuple
    return [args if type(args) == tuple else (args,) for args in params]


def _get_func_name(func):
    "Helper for finding an appropriate name for the given callable, handling ``partial`` objects."
    kw = {}
    while isinstance(func, partial):
        if func.keywords:
            kw.update(func.keywords)
        func = func.func
    funcname = func.__qualname__
    if kw:
        funcname += "(%s)" % ", ".join("%s=%r" % p for p in sorted(kw.items()))
    return funcname


def _to_log_contexts(params, log_contexts):
    "Helper for normalizing a list of parameters and log-contexts into a list of usable log context dicts"
    if not log_contexts:
        log_contexts = (dict(context=str(p) if len(p) > 1 else str(p[0])) for p in params)
    else:
        log_contexts = (p if isinstance(p, dict) else dict(context=str(p))
                        for p in log_contexts)
    return log_contexts


@contextmanager
def asynchronous(func, params=None, workers=None, log_contexts=None, final_timeout=2.0, **kw):
    """
    Map the list of tuple-parameters onto asynchronous calls to the specified function::

        with asynchronous(connect, [(host1,), (host2,), (host3,)]) as futures:
            ...

        connections = futures.results()

    :param func: The callable to invoke asynchronously.
    :param params: A list of tuples to map onto the function.
    :param workers: The number of workers to use. Defaults to the number of items in ``params``.
    :param log_contexts: A optional list of logging context objects, matching the items in ``params``.
    :param final_timeout: The amount of time to allow for the futures to complete after exiting the asynchronous context.
    """
    if params is None:
        params = [()]

    if not isinstance(params, list):  # don't use listify - we want to listify tuples too
        params = [params]

    params = _to_args_list(params)
    log_contexts = _to_log_contexts(params, log_contexts)
    if workers is None:
        workers = min(MAX_THREAD_POOL_SIZE, len(params))

    try:
        signature = inspect.signature(func)
    except ValueError:
        # In Python 3.5+, inspect.signature returns this for built-in types
        pass
    else:
        if '_sync' in signature.parameters and '_sync' not in kw:
            assert len(params) <= workers, 'SynchronizationCoordinator with %s tasks but only %s workers' % (len(params), workers)
            synchronization_coordinator = SynchronizationCoordinator(len(params))
            kw['_sync'] = synchronization_coordinator

            func = synchronization_coordinator._abandon_when_done(func)

    if not params:
        # noop
        yield Futures()
        return

    with Futures.executor(workers=workers) as futures:
        for args, ctx in zip(params, log_contexts):
            futures.submit(func, *args, log_ctx=ctx, **kw)

        yield futures


def concurrent_find(func, params, **kw):
    assert not DISABLE_CONCURRENCY, "concurrent_find runs only with concurrency enabled"
    timeout = kw.pop("concurrent_timeout", None)
    with asynchronous(func, list(params), **kw) as futures:
        future = None
        try:
            for future in futures.as_completed(timeout=timeout):
                if not future.exception() and future.result():
                    futures.kill()
                    return future.result()
            else:
                if future:
                    return future.result()
        except FutureTimeoutError as exc:
            if not timeout:
                # ??
                raise
            futures.kill()
            _logger.warning("Concurrent future timed out (%s)", exc)


def nonconcurrent_map(func, params, log_contexts=None, **kw):
    futures = Futures()
    log_contexts = _to_log_contexts(params, log_contexts)
    has_exceptions = False
    for args, ctx in zip(_to_args_list(params), log_contexts):
        future = Future()
        futures.append(future)
        try:
            result = _run_with_exception_logging(func, args, kw, ctx)
        except Exception as exc:
            has_exceptions = True
            future.set_exception(exc)
        else:
            future.set_result(result)

    if has_exceptions:
        exceptions = [f.exception() for f in futures]
        raise MultiException(exceptions=exceptions, futures=futures)

    results = [f.result() for f in futures]
    del futures[:]
    return results


def concurrent_map(func, params, workers=None, log_contexts=None, initial_log_interval=None, **kw):
    if DISABLE_CONCURRENCY or len(params) == 1:
        return nonconcurrent_map(func, params, log_contexts, **kw)

    with asynchronous(func, list(params), workers, log_contexts, **kw) as futures:
        futures.logged_wait(initial_log_interval=initial_log_interval)
        return futures.result()


# This metaclass helps generate MultiObject subtypes for specific object types

class MultiObjectMeta(type):
    _SUBTYPES = {}
    _SUBTYPES_LOCK = threading.RLock()

    def __getitem__(cls, typ):
        try:
            return cls._SUBTYPES[typ]
        except KeyError:
            with cls._SUBTYPES_LOCK:
                if typ in cls._SUBTYPES:
                    return cls._SUBTYPES[typ]
                bases = tuple(cls[base] for base in typ.__bases__ if base) or (MultiObject, )
                subtype = type("MultiObject[%s]" % typ.__qualname__, bases, dict(CONCESTOR=typ))
                cls._SUBTYPES[typ] = subtype
                return subtype

    __iter__ = None

    def __call__(cls, items=None, *args, **kwargs):
        items = tuple(items if items else [])
        common_type = concestor(*map(type, items))
        if not issubclass(common_type, cls.CONCESTOR):
            raise TypeError("%s is not a type of %s" % (common_type, cls.CONCESTOR))
        subtype = cls[common_type]
        return type.__call__(subtype, items, *args, **kwargs)


class MultiObject(object, metaclass=MultiObjectMeta):

    CONCESTOR = object

    def __init__(self, items=None, log_ctx=None, workers=None, initial_log_interval=None):
        self._items = tuple(items) if items else ()
        self._workers = workers
        self._initial_log_interval = initial_log_interval
        cstr = self.CONCESTOR
        if hasattr(cstr, '_multiobject_log_ctx'):
            # override the given log_ctx if the new items have it
            # some objects (Plumbum Cmd) are expensive to just get the attribute, so we require it
            # on the base class
            self._log_ctx = tuple(item._multiobject_log_ctx for item in self._items)
        elif callable(log_ctx):
            self._log_ctx = tuple(map(log_ctx, self._items))
        elif log_ctx:
            self._log_ctx = tuple(log_ctx)
        elif issubclass(cstr, str):
            self._log_ctx = tuple(dict(context="%s" % item) for item in self._items)
        else:
            self._log_ctx = tuple(dict(context="%s<M%03d>" % (cstr.__name__, i)) for i, item in enumerate(self._items))

        if self._workers is None and hasattr(cstr, '_multiobject_workers'):
            _workers = cstr._multiobject_workers
            if _workers == -1:
                self._workers = len(self._items) or None
            else:
                self._workers = _workers

    def __repr__(self):
        return "<%s (x%s/%s)>" % (self.__class__.__name__, len(self), self._workers)

    @property
    def L(self):
        return list(self._items)

    @property
    def T(self):
        return self._items

    @property
    def C(self):
        from .collections import ListCollection
        return ListCollection(self)

    def __getattr__(self, attr):
        if attr.startswith("_"):
            raise AttributeError()
        get = lambda obj: getattr(obj, attr)
        ret = concurrent_map(get, self, log_contexts=self._log_ctx, workers=self._workers)
        return self._new(ret)

    def __call__(self, *args, **kwargs):
        if not self:
            return self._new(self)
        for obj in self:
            if not callable(obj):
                raise Exception("%s is not callable" % obj)

        def do_it(obj, **more_kwargs):
            more_kwargs.update(kwargs)
            return obj(*args, **more_kwargs)

        if all(hasattr(obj, "__qualname__") for obj in self):
            do_it = wraps(obj)(do_it)
        else:
            common_typ = concestor(*map(type, self))
            do_it.__qualname__ = common_typ.__qualname__
        initial_log_interval = kwargs.pop("initial_log_interval", None)
        ret = concurrent_map(
            do_it, self,
            log_contexts=self._log_ctx,
            workers=self._workers,
            initial_log_interval=initial_log_interval)
        return self._new(ret)

    def __dir__(self):
        return sorted(set.intersection(*(set(dir(obj)) for obj in self)).union(super().__dir__()))

    trait_names = __dir__

    def __iter__(self):
        return iter(self._items)

    def __len__(self):
        return len(self._items)

    # ================
    def __getitem__(self, key):
        return self.call(lambda i: i[key])

    def _new(self, items=None, ctxs=None, workers=None, initial_log_interval=None):
        return MultiObject(
            self._items if items is None else items,
            self._log_ctx if ctxs is None else ctxs,
            self._workers if workers is None else workers,
            self._initial_log_interval if initial_log_interval is None else initial_log_interval)

    def with_workers(self, workers):
        "Return a new ``MultiObject`` based on current items with the specified number of workers"
        return self._new(workers=workers)

    def call(self, func, *args, **kw):
        "Concurrently call a function on each of the object contained by this ``MultiObject`` (as first param)"
        initial_log_interval = kw.pop("initial_log_interval", self._initial_log_interval)
        if kw:
            func = wraps(func)(partial(func, **kw))
        params = [((item,) + args) for item in self] if args else self
        return self._new(concurrent_map(
            func, params,
            log_contexts=self._log_ctx,
            workers=self._workers,
            initial_log_interval=initial_log_interval), initial_log_interval=initial_log_interval)

    each = call

    def filter(self, pred):
        if not pred:
            pred = bool
        filtering = self.call(pred)
        filtered = [t for (*t, passed) in zip(self, self._log_ctx, filtering) if passed]
        return self._new(*(zip(*filtered) if filtered else ((), ())))

    def chain(self):
        "Chain the iterables contained by this ``MultiObject``"
        return MultiObject(chain(*self))

    def zip_with(self, *collections):
        mo = self._new(zip(self, *collections))
        assert len(mo) == len(self), "All collection must have at least %s items" % len(self)
        return mo

    def enumerate(self, start=0):
        """
        Replaces this pattern, which loses the log contexts::

            MultiObject(enumerate(items)).call(lambda idx, item: ...)

        with this pattern, which retains log contexts::

            MultiObject(items).enumerate().call(lambda idx, item: ...)
        """
        return self._new(zip(range(start, start + len(self)), self))

    def zip(self):
        "Concurrently iterate through the iterables contained by this ``MultiObject``"
        iters = list(map(iter, self))
        while True:
            try:
                ret = concurrent_map(next, iters, log_contexts=self._log_ctx, workers=self._workers)
            except MultiException as me:
                if me.common_type == StopIteration and me.complete:
                    break
                raise
            else:
                yield self._new(ret)

    def concurrent_find(self, func=lambda f: f(), **kw):
        return concurrent_find(func, self, log_contexts=self._log_ctx, workers=self._workers, **kw)

    def __enter__(self):
        return self.call(lambda obj: obj.__enter__())

    def __exit__(self, *args):
        self.call(lambda obj: obj.__exit__(*args))


def concestor(*cls_list):
    "Closest common ancestor class"
    mros = [list(inspect.getmro(cls)) for cls in cls_list]
    track = defaultdict(int)
    while mros:
        for mro in mros:
            cur = mro.pop(0)
            track[cur] += 1
            if track[cur] == len(cls_list):
                return cur
            if len(mro) == 0:
                mros.remove(mro)
    return object  # the base-class that rules the all


class concurrent(object):
    """
    Higher-level thread execution.

    :param func: The callable to invoke asynchronously.
    :param throw: When used as a context-manager, if an exception was thrown inside the thread, re-raise it to the calling thread upon exiting the context. (default: True)
    :param daemon: Set the thread as daemon, so it does not block the process from exiting if it did not complete. (default: True)
    :param threadname: Set a name for this thread. (default: ``anon-<id>``)

    :param loop: If ``True``, repeatedly calls ``func`` until the context is exited, ``.stop()`` is called, or the ``stopper`` event object is set. (default: False)
    :param sleep: Used with the ``loop`` flag - the number of seconds between consecutive calls to ``func``. (default: 1)
    :param stopper: Used with the ``loop`` flag - an external ``threading.Event`` object to use for stopping the loop .

    :param console_logging: If ``False``, suppress logging to the console log handler. (default: False)

    Running multiple tasks side-by-side::

        with \
            concurrent(requests.get, "api.myserver.com/data1") as async1, \
            concurrent(requests.get, "api.myserver.com/data2") as async2:
            my_data = open("local_data").read()

        remote_data1 = async1.result()
        remote_data2 = async2.result()
        process_it(my_data, remote_data1, remote_data2)

    Run something repeatedly in the background:

        heartbeats = concurrent(send_heartbeat, loop=True, sleep=5)
        heartbeats.start()  # until stopped, or the process exits

        with concurrent(print_status, loop=True, sleep=5):
            some_long_task()

        heartbeats.stop()
    """

    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.throw = kwargs.pop('throw', True)
        self.daemon = kwargs.pop('daemon', True)
        self.stopper = kwargs.pop('stopper', threading.Event())
        self.sleep = kwargs.pop('sleep', 1)
        self.loop = kwargs.pop('loop', False)
        self.timer = None
        self.console_logging = kwargs.pop('console_logging', True)
        self.threadname = kwargs.pop('threadname', None)
        if not self.threadname:
            current_thread_name = threading.current_thread().name
            if current_thread_name:
                current_thread_name = current_thread_name.split('::')[0]  # We want to see only the context
                self.threadname = "%s::%X" % (current_thread_name, id(self))
            else:
                self.threadname = "anon-%X" % id(self)

        real_thread_no_greenlet = kwargs.pop('real_thread_no_greenlet', False)
        if IS_GEVENT:
            # in case of using apply_gevent_patch function - use this option in order to defer some jobs to real threads
            self.real_thread_no_greenlet = real_thread_no_greenlet
        else:
            # gevent isn't active, no need to do anything special
            self.real_thread_no_greenlet = False

        rimt = kwargs.pop("raise_in_main_thread", False)
        if rimt:
            exc_type = Exception if rimt is True else rimt
            self.func = raise_in_main_thread(exc_type)(self.func)

    def __repr__(self):
        flags = ""
        if self.daemon:
            flags += 'D'
        if self.loop:
            flags += 'L'
        if self.real_thread_no_greenlet:
            flags += 'T'
        return "<%s[%s] '%s'>" % (self.__class__.__name__, self.threadname, flags)

    def _logged_func(self):
        stack = ExitStack()
        self.exc = None
        self.timer = Timer()
        stack.callback(self.timer.stop)
        stack.callback(self.stop)
        try:
            if not self.console_logging:
                stack.enter_context(_logger.suppressed())
            _logger.debug("%s - starting", self)
            while True:
                self._result = self.func(*self.args, **self.kwargs)
                if not self.loop:
                    return
                if self.wait(self.sleep):
                    _logger.debug("%s - stopped", self)
                    return
        except ProcessExiting as exc:
            _logger.debug(exc)
            raise
        except KeyboardInterrupt as exc:
            _logger.silent_exception("KeyboardInterrupt in thread running %s:", self.func)
            self.exc = apply_timestamp(exc)
            if IS_GEVENT:
                raise  # in gevent we should let this exception propagate to the main greenlet
        except Exception as exc:
            _logger.silent_exception("Exception in thread running %s: %s (traceback can be found in debug-level logs)", self.func, type(exc))
            self.exc = apply_timestamp(exc)
        finally:
            stack.close()

    def stop(self):
        _logger.debug("%s - stopping", self)
        self.stopper.set()

    def wait(self, timeout=None):
        if self.real_thread_no_greenlet:
            # we can't '.wait' on this gevent event object, so instead we test it and sleep manually:
            timer = Timer(expiration=timeout)
            while not timer.expired:
                if self.stopper.is_set():
                    return True
                non_gevent_sleep(0.1)
            return False
        return self.stopper.wait(timeout)

    def result(self, timeout=None):
        self.wait(timeout=timeout)
        if self.throw and self.exc:
            raise self.exc
        return self._result

    def done(self):
        """
        Return ``True`` if the thread is done (successfully or not)
        """
        return hasattr(self, '_result') or getattr(self, 'exc', None) is not None

    @contextmanager
    def paused(self):
        self.stop()
        yield
        self.start()

    @contextmanager
    def _running(self):
        if DISABLE_CONCURRENCY:
            self._logged_func()
            yield self
            return

        if self.real_thread_no_greenlet and IS_GEVENT:
            _logger.debug('sending job to a real OS thread')
            self._join = defer_to_thread(func=self._logged_func, threadname=self.threadname)
        else:
            # threading.Thread could be a real thread or a gevent-patched thread...
            self.thread = threading.Thread(target=self._logged_func, name=self.threadname, daemon=self.daemon)
            _logger.debug('sending job to %s', self.thread)
            self.stopper.clear()
            self.thread.start()
            self._join = self.thread.join
        try:
            yield self
        finally:
            self.stop()  # if we loop, stop it
        self._join()
        if self.throw and self.exc:
            raise self.exc

    def __enter__(self):
        self._ctx = self._running()
        return self._ctx.__enter__()

    def __exit__(self, *args):
        return self._ctx.__exit__(*args)

    def __iter__(self):
        # TODO: document or remove
        with self:
            self.iterations = 0
            while not self.wait(self.sleep):
                yield self
                self.iterations += 1

    start = __enter__

    def join(self):
        self.__exit__(None, None, None)

    __del__ = join


# re-exports
from .sync import break_locks, TerminationSignal, initialize_exception_listener, initialize_termination_listener, Timebomb
from .sync import set_timebomb, TagAlongThread, SYNC, LoggedRLock, RWLock, SoftLock, skip_if_locked, with_my_lock
from .sync import synchronized, SynchronizedSingleton, LoggedCondition, _check_exiting


import sys
if sys.version_info < (3, 7):
    # async became reserved in 3.7, but we'll keep it for backwards compatibility
    code = compile("async = asynchronous", __name__, "exec")
    eval(code, globals(), globals())



# encoding: utf-8
from __future__ import absolute_import

import logging
import os
import random
import sys
import threading
import time
import traceback
from contextlib import ExitStack
from functools import wraps, partial
from itertools import cycle, chain, repeat, count
from collections import OrderedDict

from easypy.colors import colorize, uncolored
from easypy.humanize import compact as _compact
from easypy.timing import timing as timing_context, Timer
from easypy.threadtree import ThreadContexts
from easypy.contexts import contextmanager

logging.INFO1 = logging.INFO+1
logging.addLevelName(logging.INFO1, "INFO1")


CLEAR_EOL = '\x1b[0K'
IS_A_TTY = sys.stdout.isatty()

LOG_WITH_COLOR=None
def compact(msg):
    raise NotImplementedError()

def set_width(TERM_WIDTH):
    if TERM_WIDTH in (True, None):
        TERM_WIDTH, _ = os.get_terminal_size()
        TERM_WIDTH = max(TERM_WIDTH, 120)
    if TERM_WIDTH:
        compact = lambda line: _compact(line, TERM_WIDTH-5)
    else:
        TERM_WIDTH = 0
        compact = lambda line: line
    globals().update(locals())


def set_coloring(enabled):
    global LOG_WITH_COLOR
    LOG_WITH_COLOR = enabled
    if enabled:
        from easypy.colors import RED, GREEN, BLUE, WHITE, DARK_GRAY
    else:
        RED = GREEN = BLUE = WHITE = DARK_GRAY = lambda txt, *_, **__: txt
    globals().update(locals())


# All this expected to be set by set_graphics call
LINE = None
DOUBLE_LINE = None
INDENT_SEGMENT = None
INDENT_OPEN = None
INDENT_CLOSE = None
INDENT_EXCEPTION = None

def set_graphics(GRAPHICAL):
    if GRAPHICAL:
        LINE = "─"
        DOUBLE_LINE = "═"
        INDENT_SEGMENT   = "  │ "
        INDENT_OPEN      = "  ├───┮ "
        INDENT_CLOSE         = "  ╰╼"
        INDENT_EXCEPTION     = "  ╘═"
    else:
        LINE = "-"
        DOUBLE_LINE = "="
        INDENT_SEGMENT   = "..| "
        INDENT_OPEN      = "..|---+ "
        INDENT_CLOSE         = "  '-"
        INDENT_EXCEPTION     = "  '="
    globals().update(locals())

set_width(IS_A_TTY)
set_coloring(IS_A_TTY or os.environ.get('TERM_COLOR_SUPPORT'))
set_graphics(IS_A_TTY or os.environ.get('TERM_COLOR_SUPPORT'))


LEVEL_COLORS = {
    logging.DEBUG:   "DARK_GRAY",
    logging.INFO:    "GRAY",
    logging.WARNING: "YELLOW",
    logging.ERROR:   "RED"
    }


def get_level_color(level):
    try:
        return LEVEL_COLORS[level]
    except KeyError:
        sorted_colors = sorted(LEVEL_COLORS.items(), reverse=True)
        for clevel, color in sorted_colors:
            if level > clevel:
                break
        LEVEL_COLORS[level] = color
        return color

INDENT_COLORS = [
    ("DARK_%s<<{}>>" % color.upper()).format
    for color in "GREEN BLUE MAGENTA CYAN YELLOW".split()]
random.shuffle(INDENT_COLORS)


class LogLevelClamp(logging.Filterer):
    def __init__(self, level=logging.DEBUG):
        self.level = level
        self.name = logging.getLevelName(level)
    def filter(self, record):
        if record.levelno > self.level:
            record.levelname, record.levelno = self.name, self.level
        return True


def get_console_handler():
    try:
        return logging._handlers['console']
    except KeyError:
        for handler in logging.root.handlers:
            if not isinstance(handler, logging.StreamHandler):
                continue
            return handler


class ThreadControl(logging.Filter):
    """
    Used by ContextLoggerMixin .solo and .suppressed methods to control logging to console
    To use, add it to the logging configuration as a filter in the console handler

        ...
        'filters': {
            'thread_control': {
                '()': 'easypy.logging.ThreadControl'
            }
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'filters': ['thread_control'],
            },

    """

    CONTEXT = ThreadContexts(counters='silenced')

    # we use this ordered-dict to track which thread is currently 'solo-ed'
    # we populate it with some initial values to make the 'filter' method
    # implementation more convenient
    SELECTED = OrderedDict()
    IDX_GEN = count()
    LOCK = threading.RLock()

    @classmethod
    @contextmanager
    def solo(cls):
        try:
            with cls.LOCK:
                idx = next(cls.IDX_GEN)
                cls.SELECTED[idx] = threading.current_thread()
            yield
        finally:
            cls.SELECTED.pop(idx)

    def filter(self, record):
        selected = False
        while selected is False:
            idx = next(reversed(self.SELECTED), None)
            if idx is None:
                selected = None
                break
            selected = self.SELECTED.get(idx, False)

        if selected:
            return selected == threading.current_thread()
        return not self.CONTEXT.silenced


class ConsoleFormatter(logging.Formatter):
    def formatMessage(self, record):
        if not hasattr(record, "levelcolor"):
            record.levelcolor = get_level_color(record.levelno)
        msg = super().formatMessage(record)
        if IS_A_TTY:
            msg = '\r' + msg + CLEAR_EOL
        return colorize(msg) if LOG_WITH_COLOR else uncolored(msg)


try:
    import yaml
except ImportError:
    pass
else:
    try:
        from yaml import CDumper as Dumper
    except ImportError:
        from yaml import Dumper

    class YAMLFormatter(logging.Formatter):

        def __init__(self, **kw):
            self.dumper_params = kw

        def format(self, record):
            return yaml.dump(vars(record), Dumper=Dumper) + "\n---\n"


def configure_contextual_logging(_ctx=ExitStack(), **kw):
    indentation = int(os.getenv("EASYPY_LOG_INDENTATION", "0"))
    _ctx.enter_context(THREAD_LOGGING_CONTEXT(indentation=indentation, **kw))


THREAD_LOGGING_CONTEXT = ThreadContexts(counters="indentation", stacks="context", defaults=dict(host=''))
get_current_context = THREAD_LOGGING_CONTEXT.flatten


def get_indentation():
    return THREAD_LOGGING_CONTEXT.indentation


def _progress():
    from random import randint
    while True:
        yield chr(randint(0x2800, 0x28FF))


class ProgressBar:

    WAITING = "▅▇▆▃  ▆▇▆▅▃_       " #
    #PROGRESSING = "⣾⣽⣻⢿⡿⣟⣯⣷" #"◴◷◶◵◐◓◑◒"
    SPF = 1.0/15

    def __init__(self):
        self._event = threading.Event()
        self._thread = None
        self._lock = threading.RLock()
        self._depth = 0

    def loop(self):
        wait_seq = cycle(self.WAITING)
        prog_seq = _progress()
        wait_symb, progress_symb = map(next, (wait_seq, prog_seq))
        last_time = hanging = 0
        while True:
            progressed = self._event.wait(self.SPF)
            if self._stop:
                break
            now = time.time()

            if now - last_time >= self.SPF:
                wait_symb = next(wait_seq)
                last_time = now

            if progressed:
                progress_symb = next(prog_seq)
                hanging = 0
            else:
                hanging +=1

            anim = WHITE(wait_symb+progress_symb)

            elapsed = self._timer.elapsed.render(precision=0).rjust(8)
            if hanging >= (5*10*60):  # ~5 minutes with no log messages
                elapsed = RED(elapsed)
            else:
                elapsed = BLUE(elapsed)

            line = elapsed + self._last_line.rstrip()
            line = line.replace("__PB__", anim)
            print("\r" + line, end=CLEAR_EOL+"\r", flush=True)
            self._event.clear()
        print("\rDone waiting.", end=CLEAR_EOL+"\r", flush=True)

    def progress(self, record):
        if not self._thread:
            return
        if record.levelno >= logging.DEBUG:
            record.drawing = "__PB__" + record.drawing[2:]
            self._last_line = compact(uncolored(self._format(record).split("\n")[0]).strip()[8:])
        self._event.set()

    def set_message(self, msg):
        msg = msg.replace("|..|", "|__PB__"+INDENT_SEGMENT[3])
        self._last_line = "|" + compact(msg)
        self._event.set()

    @contextmanager
    def __call__(self):
        if not GRAPHICAL:
            yield self
            return

        handler = get_console_handler()
        with self._lock:
            self._depth += 1
            if self._depth == 1:
                self.set_message("Waiting...")
                self._stop = False
                self._timer = Timer()
                self._format = handler.formatter.format if handler else lambda record: record.getMessage()
                self._thread = threading.Thread(target=self.loop, name="ProgressBar", daemon=True)
                self._thread.start()
        try:
            yield self
        finally:
            with self._lock:
                self._depth -= 1
                if self._depth <= 0:
                    self._stop = True
                    self._event.set()
                    self._thread.join()
                    self._thread = None


class ProgressHandler(logging.NullHandler):
    def handle(self, record):
        PROGRESS_BAR.progress(record)


PROGRESS_BAR = ProgressBar()


class AbortedException(BaseException):
    """ Aborted base class

    Exceptions that inherit from this class will show as ABORTED in logger.indented
    """


class ContextLoggerMixin(object):

    _debuggifier = LogLevelClamp()

    @contextmanager
    def context(self, context=None, indent=False, progress_bar=False, **kw):
        if context:
            kw['context'] = context
        with ExitStack() as stack:
            stack.enter_context(THREAD_LOGGING_CONTEXT(kw))
            timing = kw.pop("timing", True)
            if indent:
                header = indent if isinstance(indent, str) else ("[%s]" % context)
                stack.enter_context(self.indented(header=header, timing=timing))
            if progress_bar:
                stack.enter_context(self.progress_bar())
            yield

    def suppressed(self):
        """
        Context manager - Supress all logging to the console from the calling thread
        """
        return ThreadControl.CONTEXT(silenced=True)

    def solo(self):
        """
        Context manager - Allow logging to the console from the calling thread only
        """
        return ThreadControl.solo()

    @contextmanager
    def indented(self, header=None, *args, level=logging.INFO1, timing=True, footer=True):
        header = compact((header % args) if header else "")
        self._log(level, "WHITE@{%s}@" % header, (), extra=dict(drawing=INDENT_OPEN))
        with ExitStack() as stack:
            stack.enter_context(THREAD_LOGGING_CONTEXT(indentation=1))

            get_duration = lambda: ""
            if timing:
                timer = stack.enter_context(timing_context())
                get_duration = lambda: " in DARK_MAGENTA<<{:text}>>".format(timer.duration)

            def footer_log(color, title, drawing):
                if footer:
                    self._log(level, "%s@{%s}@%s (%s)", (color, title, get_duration(), header), extra=dict(drawing=drawing))
                else:
                    self._log(level, "", (), extra=dict(drawing=drawing))

            try:
                yield
            except (KeyboardInterrupt, AbortedException):
                footer_log("CYAN", "ABORTED", INDENT_EXCEPTION)
                raise
            except GeneratorExit:
                footer_log("DARK_GRAY", "DONE", INDENT_CLOSE)
            except:
                footer_log("RED", "FAILED", INDENT_EXCEPTION)
                raise
            else:
                footer_log("DARK_GRAY", "DONE", INDENT_CLOSE)

    def error_box(self, *exc, extra=None):
        if len(exc)==1:
            exc, = exc
            typ = type(exc)
            tb = None
        else:
            typ, exc, tb = exc
        header = "%s.%s" % (typ.__module__, typ.__name__)
        self.error("YELLOW@{%s}@ RED@{%s}@", header, LINE*(80-len(header)-1), extra=dict(drawing=RED(INDENT_OPEN)))
        with THREAD_LOGGING_CONTEXT(indentation=1, drawing=RED(INDENT_SEGMENT)):
            if hasattr(exc, "render") and callable(exc.render):
                exc_text = exc.render()
            elif tb:
                fmt = "DARK_GRAY@{{{}}}@"
                full_traceback = "".join(traceback.format_exception(typ, exc, tb))
                exc_text = "\n".join(map(fmt.format, full_traceback.splitlines()))
            else:
                exc_text = str(exc)
            for line in exc_text.splitlines():
                self.error(line)
            if extra:
                for line in extra.splitlines():
                    self.error(line)
            self.error("RED@{%s}@", DOUBLE_LINE*80, extra=dict(drawing=RED(INDENT_EXCEPTION)))

    _progressing = False
    @contextmanager
    def progress_bar(self):
        if not GRAPHICAL:
            with PROGRESS_BAR() as pb:
                yield pb
                return

        with ExitStack() as stack:
            if not self.__class__._progressing:
                self.addFilter(self._debuggifier)
                stack.callback(self.removeFilter, self._debuggifier)
                stack.enter_context(PROGRESS_BAR())
                self.__class__._progressing = True
                stack.callback(setattr, self.__class__, "_progressing", False)
            yield PROGRESS_BAR

    def silent_exception(self, message, *args, **kwargs):
        "like ``exception()``, only emits the traceback in debug level"
        self.error(message, *args, **kwargs)
        self.debug('Traceback:', exc_info=True)

    def __rand__(self, cmd):
        return cmd & self.pipe(logging.INFO, logging.INFO)

    def pipe(self, err_level=logging.DEBUG, out_level=logging.INFO, prefix=None, line_timeout=10 * 60, **kw):
        class LogPipe(object):
            def __rand__(_, cmd):
                popen = cmd if hasattr(cmd, "iter_lines") else cmd.popen()
                for out, err in popen.iter_lines(line_timeout=line_timeout, **kw):
                    for level, line in [(out_level, out), (err_level, err)]:
                        if not line:
                            continue
                        for l in line.splitlines():
                            if prefix:
                                l = "%s: %s" % (prefix, l)
                            self.log(level, l)
                return popen.returncode
        return LogPipe()

    def pipe_info(self, prefix=None, **kw):
        return self.pipe(logging.INFO, logging.INFO, prefix=prefix, **kw)

    def pipe_debug(self, prefix=None, **kw):
        return self.pipe(logging.DEBUG, logging.DEBUG, prefix=prefix, **kw)

    def info1(self, *args, **kwargs):
        return self.log(logging.INFO1, *args, **kwargs)

    def announced_vars(self, header='With locals:', *args, **kwargs):
        "Announces the variables declared in the context"
        import inspect
        frame = inspect.currentframe().f_back

        # `@contextmanager` annotates an internal `cm` function instead of the
        # `announced_vars` method so that `inspect.currentframe().f_back` will
        # point to the frame that uses `announced_vars`. If we decoraed
        # `announced_vars` with `@contextmanager`, we'd have to depend on
        # implementation details of `@contextmanager` - currently
        # `inspect.currentframe().f_back.f_back` would have worked, but we have
        # no guarantee that it'll remain like this forever.
        @contextmanager
        def cm():
            old_local_names = set(frame.f_locals.keys())
            yield
            new_locals = frame.f_locals
            with ExitStack() as stack:
                if header:
                    stack.enter_context(self.indented(header, *args, footer=False, **kwargs))
                # Traverse co_varnames to retain order
                for name in frame.f_code.co_varnames:
                    if name not in old_local_names and name in new_locals:
                        self.info('%s = %s', name, new_locals[name])

                # Print the names we somehow missed(because they weren't in co_varnames - it can happen!)
                for name in (new_locals.keys() - old_local_names - set(frame.f_code.co_varnames)):
                    self.info('%s = %s', name, new_locals[name])

        return cm()


if not issubclass(logging.Logger, ContextLoggerMixin):
    logging.Logger.__bases__ = logging.Logger.__bases__ + (ContextLoggerMixin,)

    def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
        drawing = INDENT_SEGMENT

        rv = logging.Logger._makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=func, sinfo=sinfo)
        if extra is not None:
            drawing = extra.pop('drawing', drawing)
            for key in extra:
                if (key in ["message", "asctime"]) or (key in rv.__dict__):
                    raise KeyError("Attempt to overwrite %r in LogRecord" % key)
                rv.__dict__[key] = extra[key]

        contexts = THREAD_LOGGING_CONTEXT.context
        extra = THREAD_LOGGING_CONTEXT.flatten()
        extra['context'] = "[%s]" % ";".join(contexts) if contexts else ""
        rv.__dict__.update(dict(extra, **rv.__dict__))

        indents = chain(repeat(INDENT_SEGMENT, rv.indentation), repeat(drawing, 1))
        rv.drawing = "".join(color(segment) for color, segment in zip(cycle(INDENT_COLORS), indents))
        return rv

    logging.Logger._makeRecord, logging.Logger.makeRecord = logging.Logger.makeRecord, makeRecord


class HeartbeatHandler(logging.Handler):
    "Heartbeat notifications based on the application's logging activity"

    def __init__(self, beat_func, min_interval=1, **kw):
        """
        @param beat_func: calls this function when a heartbeat is due
        @param min_interval: minimum time interval between heartbeats
        """
        super(HeartbeatHandler, self).__init__(**kw)
        self.min_interval = min_interval
        self.last_beat = 0
        self.beat = beat_func
        self._emitting = False

    def emit(self, record):
        if self._emitting:
            # prevent reenterance
            return

        try:
            self._emitting = True
            if (record.created - self.last_beat) > self.min_interval:
                try:
                    log_message = self.format(record)
                except:
                    log_message = "Log record formatting error (%s:#%s)" % (record.filename, record.lineno)
                self.beat(log_message=log_message, heartbeat=record.created)
                self.last_beat = record.created
        finally:
            self._emitting = False


def log_context(method=None, **ctx):
    if not method:
        return partial(log_context, **ctx)
    @wraps(method)
    def inner(*args, **kwargs):
        context = {k: fmt.format(*args, **kwargs) for k, fmt in ctx.items()}
        with THREAD_LOGGING_CONTEXT(context):
            return method(*args, **kwargs)
    return inner


#=====================#=====================#=====================#
# This monkey-patch tricks logging's findCaller into skipping over
# this module when looking for the caller of a logger.log function

try:
    # restore, in case we've already mocked this, as when running unit-tests
    logging._srcfile = logging._orig_srcfile
except AttributeError:
    logging._orig_srcfile = logging._srcfile


class _SrcFiles:
    _srcfiles = {logging._srcfile, __file__}

    def __eq__(self, fname):
        return fname in self.__class__._srcfiles


logging._srcfile = _SrcFiles()
#=====================#=====================#=====================#



_root = __file__[:__file__.find(os.sep.join(__name__.split(".")))]

def _trim(pathname, modname, cache={}):
    try:
        return cache[(pathname, modname)]
    except KeyError:
        pass

    elems = pathname.replace(_root, "").strip(".").split(os.sep)[:-1]
    if modname != "__init__":
        elems.append(modname)

    ret = cache[(pathname, modname)] = filter(None, elems)
    return ret



import pytest
from easypy.aliasing import aliases


def test_aliasing_static():

    @aliases("this")
    class Foo():
        this = dict(a=1)

    f = Foo()
    assert f.get("a") == 1


def test_aliasing_dynamic():

    class Foo():
        def __init__(self):
            self.this = dict(a=1)

    with pytest.raises(AssertionError):
        Foo = aliases("this")(Foo)

    Foo = aliases("this", static=False)(Foo)

    f = Foo()
    assert f.get("a") == 1


def test_aliasing_inherit():

    class Foo(int):
        def __init__(self, x):
            self.this = dict(a=1)

    with pytest.raises(AssertionError):
        Foo = aliases("this")(Foo)

    Foo = aliases("this", static=False)(Foo)

    f = Foo("5")
    assert f.get("a") == 1
    assert f == 5


def test_aliasing_infinite_recursion_exception():
    @aliases('bar', static=False)
    class Foo:
        def __init__(self):
            self.bar = Bar(self)

        def __repr__(self):
            return 'Foo()'

    class Bar:
        def __init__(self, foo):
            self.foo = foo

        def baz(self):
            return self.foo.baz()

    with pytest.raises(getattr(__builtins__, 'RecursionError', RuntimeError)) as e:
        Foo().baz()
    assert str(e.value) == "Infinite recursion trying to access 'baz' on Foo() (via Foo.bar.baz)"



import os
import time
from logging import getLogger
from uuid import uuid4
import random
import weakref
import gc

import pytest

from easypy.bunch import Bunch
from easypy.caching import timecache, PersistentCache, cached_property, locking_cache
from easypy.units import DAY
from easypy.resilience import resilient

_logger = getLogger(__name__)


def test_timecache():
    ts = 0
    data = Bunch(a=0, b=0)

    def get_ts():
        return ts

    @timecache(expiration=1, get_ts_func=get_ts, key_func=lambda k: k)
    def inc(k, x):
        x += 1
        data[k] += 1

    assert data.a == data.b == 0
    inc('a', random.random())
    assert (data.a, data.b) == (1, 0)

    inc('a', x=random.random())
    assert (data.a, data.b) == (1, 0)

    ts += 1
    inc('a', random.random())
    assert (data.a, data.b) == (2, 0)

    inc('b', x=random.random())
    assert (data.a, data.b) == (2, 1)

    inc('b', random.random())
    assert (data.a, data.b) == (2, 1)

    ts += 1
    inc('b', x=random.random())
    assert (data.a, data.b) == (2, 2)

    inc.cache_clear()
    inc('a', x=random.random())
    assert (data.a, data.b) == (3, 2)
    inc('b', x=random.random())
    assert (data.a, data.b) == (3, 3)

    inc.cache_clear()
    inc('a', x=random.random())
    inc('b', x=random.random())
    inc('a', x=random.random())
    inc('b', x=random.random())
    assert (data.a, data.b) == (4, 4)
    inc.cache_pop('a', x=random.random())
    inc('a', x=random.random())
    inc('b', x=random.random())


def test_timecache_method():
    ts = 0

    def get_ts():
        return ts

    class Foo:
        def __init__(self, prefix):
            self.prefix = prefix

        @timecache(expiration=1, get_ts_func=get_ts, key_func=lambda args: args)
        def foo(self, *args):
            return [self.prefix] + list(args)

    foo1 = Foo(1)
    foo2 = Foo(2)

    assert foo1.foo(1, 2, 3) == foo1.foo(1, 2, 3)
    assert foo1.foo(1, 2, 3) != foo1.foo(1, 2, 4)
    assert foo1.foo(1, 2, 3) != foo2.foo(1, 2, 3)

    foo1_1 = foo1.foo(1)
    foo1_2 = foo1.foo(2)
    foo2_1 = foo2.foo(1)
    foo2_2 = foo2.foo(2)

    assert foo1_1 == [1, 1]
    assert foo1_2 == [1, 2]
    assert foo2_1 == [2, 1]
    assert foo2_2 == [2, 2]

    assert foo1_1 is foo1.foo(1)
    assert foo1_2 is foo1.foo(2)
    assert foo2_1 is foo2.foo(1)
    assert foo2_2 is foo2.foo(2)

    assert foo1_1 is foo1.foo(1)
    assert foo1_2 is foo1.foo(2)
    assert foo2_1 is foo2.foo(1)
    assert foo2_2 is foo2.foo(2)

    foo1.foo.cache_clear()
    foo2.foo.cache_pop(1)

    assert foo1_1 is not foo1.foo(1)
    assert foo1_2 is not foo1.foo(2)
    assert foo2_1 is not foo2.foo(1)
    assert foo2_2 is foo2.foo(2)


def test_timecache_getattr():
    ts = 0

    def get_ts():
        return ts

    class Foo:
        def __init__(self):
            self.count = 0

        @timecache(expiration=1, get_ts_func=get_ts)
        def __getattr__(self, name):
            self.count += 1
            return [self.count, name]

    foo = Foo()

    assert foo.bar == [1, 'bar']
    assert foo.bar == [1, 'bar']
    assert foo.baz == [2, 'baz']

    ts += 1

    assert foo.baz == [3, 'baz']
    assert foo.bar == [4, 'bar']


@pytest.yield_fixture()
def persistent_cache_path():
    cache_path = '/tmp/test_pcache_%s' % uuid4()
    try:
        yield cache_path
    finally:
        try:
            os.unlink("%s.db" % cache_path)
        except:  # noqa
            pass


def test_persistent_cache(persistent_cache_path):
    ps = PersistentCache(persistent_cache_path, version=1)
    TEST_KEY = "test_key"
    TEST_VALUE = "test_value"
    ps.set(TEST_KEY, TEST_VALUE)
    assert ps.get(TEST_KEY) == TEST_VALUE, "Value does not match set value"

    ps = PersistentCache(persistent_cache_path, version=1)
    assert ps.get(TEST_KEY) == TEST_VALUE, "Value does not match set value after reopen"

    ps = PersistentCache(persistent_cache_path, version=2)
    with pytest.raises(KeyError):  # Changed version should invalidate cache
        ps.get(TEST_KEY)

    # Default values
    assert ps.get(TEST_KEY, default=None) is None, "Wrong default value returnen(not None)"
    assert ps.get(TEST_KEY, default="1") == "1", "Wrong default value returned"

    # Cached func should be called only once
    value_generated = False
    use_cache = True

    class UnnecessaryFunctionCall(Exception):
        pass

    ps = PersistentCache(persistent_cache_path, version=2, ignored_keywords="x")

    @ps(validator=lambda _, **__: use_cache)
    def cached_func(x):
        nonlocal value_generated
        if value_generated:
            raise UnnecessaryFunctionCall()
        value_generated = True
        return True

    assert cached_func(x=random.random()) is cached_func(x=random.random())
    assert value_generated

    # Testing validator
    use_cache = False
    with pytest.raises(UnnecessaryFunctionCall):
        cached_func(x=random.random())

    # Removing data
    ps.clear()
    assert ps.get(TEST_KEY, default=None) is None, "Database was not cleared properly"

    # Expiration
    ps = PersistentCache(persistent_cache_path, version=3, expiration=.01)
    ps.set(TEST_KEY, TEST_VALUE)
    time.sleep(0.011)
    assert ps.get(TEST_KEY, None) is None, "Database was not cleaned up on expiration"


def test_locking_timecache():
    from easypy.concurrency import MultiObject

    # Cached func should be called only once
    value_generated = False

    class UnnecessaryFunctionCall(Exception):
        pass

    @timecache(ignored_keywords='x')
    def test(x):
        nonlocal value_generated
        if value_generated:
            raise UnnecessaryFunctionCall()
        value_generated = True
        return True

    MultiObject(range(10)).call(lambda x: test(x=x))


@pytest.mark.parametrize('cache_decorator', [cached_property, timecache()])
def test_caching_gc_leaks(cache_decorator):
    """
    Make sure that the cache does not prevent GC collection once the original objects die
    """

    class Leaked():
        pass

    class Foo:
        @cache_decorator
        def cached_method(self):
            return Leaked()

        def get(self):
            """Generalize property type and function type caches"""
            result = self.cached_method
            if callable(result):
                result = result()
            assert isinstance(result, Leaked), 'cache not used properly - got wrong value %s' % (result,)
            return result

    foo = Foo()
    leaked = weakref.ref(foo.get())

    gc.collect()
    assert leaked() == foo.get()

    del foo
    gc.collect()
    assert leaked() is None


def test_resilient_between_timecaches():
    class ExceptionLeakedThroughResilient(Exception):
        pass

    @timecache(1)
    @resilient(acceptable=ExceptionLeakedThroughResilient, default='default')
    @timecache(1)
    def foo():
        raise ExceptionLeakedThroughResilient()

    assert foo() == 'default'



from easypy.colors import Colorized
from easypy.collections import listify


class CancelledException(KeyboardInterrupt):
    pass


def message(fmt, *args, wait_for_user=False, **kwargs):
    msg = fmt.format(*args, **kwargs) if (args or kwargs) else fmt
    print(Colorized(msg))
    if wait_for_user:
        globals()['wait_for_user']()


def get_input(prompt, default=NotImplemented):
    prompt = prompt.strip(": ")
    if default is not NotImplemented:
        prompt = "%s DARK_CYAN<<[%s]>>" % (prompt, default)
    ret = input(Colorized(prompt + ": "))
    if not ret:
        if default is NotImplemented:
            raise CancelledException()
        else:
            return default
    return ret


def wait_for_user():
    try:
        get_input("(Hit <enter> to continue)")
    except CancelledException:
        pass


def choose(question, options, default=NotImplemented):
    options = list(options.items() if isinstance(options, dict) else options)

    all_options = {}
    for opt, value in options:
        for o in listify(opt):
            all_options[o.lower()] = value

    disp_options = "/".join(str(opt[0] if isinstance(opt, (list, tuple)) else opt) for opt, value in options)
    fmt = "%s (%s) " if default is NotImplemented else "%s [%s]"
    msg = fmt % (question, disp_options)

    while True:
        answer = get_input(msg, default=default).strip().lower()
        if answer not in all_options:
            print("Invalid answer ('%s')" % answer)
            continue
        return all_options[answer]


def ask(question, default=NotImplemented):
    if default is NotImplemented:
        options = [(("y", "yes"), True), (("n", "no"), False)]  # no default
    elif default:
        default = "Y"
        options = [(("Y", "yes"), True), (("n", "no"), False)]  # 'Y' is default
    else:
        default = "N"
        options = [(("y", "yes"), True), (("N", "no"), False)]  # 'N' is default

    try:
        return choose(question, options, default=default)
    except EOFError:
        return False  # EOF == no



"""
This module monkey-patches python's (3.x) threading module so that we can find which thread spawned which.
This is useful in our contextualized logging, so that threads inherit logging context from their 'parent'.
"""
import sys
from importlib import import_module
from uuid import uuid4
from weakref import WeakKeyDictionary
import time
import logging
from copy import deepcopy
from contextlib import contextmanager
from logging import getLogger
import _thread
import threading

from easypy.humanize import format_thread_stack

from easypy.gevent import main_thread_ident_before_patching, is_module_patched
from .bunch import Bunch
from .collections import ilistify

from ._multithreading_init import UUIDS_TREE, IDENT_TO_UUID, UUID_TO_IDENT, MAIN_UUID
_logger = getLogger(__name__)


def get_thread_uuid(thread=None):
    """
    Assigns and returns a UUID to our thread, since thread.ident can be recycled.
    The UUID is used in mapping child threads to their parent threads.
    """
    if not thread:
        thread = threading.current_thread()

    ident = thread.ident
    try:
        uuid = IDENT_TO_UUID[ident]
    except KeyError:
        uuid = IDENT_TO_UUID.setdefault(ident, uuid4())
        UUID_TO_IDENT[uuid] = ident
    return uuid


_REGISTER_GREENLETS = False
_orig_start_new_thread = _thread.start_new_thread


def start_new_thread(target, *args, **kwargs):
    """
    A wrapper for the built in 'start_new_thread' used to capture the parent of each new thread.
    """
    parent_uuid = get_thread_uuid()

    def wrapper(*args, **kwargs):
        thread = threading.current_thread()
        uuid = get_thread_uuid(thread)
        UUIDS_TREE[uuid] = parent_uuid
        if _REGISTER_GREENLETS:
            IDENT_TO_GREENLET[thread.ident] = gevent.getcurrent()
        try:
            return target(*args, **kwargs)
        finally:
            IDENT_TO_UUID.pop(thread.ident)

    return _orig_start_new_thread(wrapper, *args, **kwargs)


get_parent_uuid = UUIDS_TREE.get


def get_thread_parent(thread):
    """
    Returns parent thread for the given thread.
    If the parent thread died, returns a ``DeadThread`` object, to preserve the thread-tree structure.
    """
    uuid = thread.uuid
    parent_uuid = UUIDS_TREE.get(uuid)
    if not parent_uuid:
        return

    # double check that values synchronized while locked
    with threading._active_limbo_lock:
        parent_ident = UUID_TO_IDENT.get(parent_uuid)
        uuid_at_ident = IDENT_TO_UUID.get(parent_ident)
        if uuid_at_ident != parent_uuid:
            # a new thread assumed this ident[ity], which is not the original parent
            return DeadThread.get(parent_uuid)

        return threading._active.get(parent_ident) or DeadThread.get(parent_uuid)


_thread.start_new_thread = start_new_thread
threading._start_new_thread = start_new_thread


class DeadThread(object):
    """
    A dummy object representing a thread that already died.
    """

    name = "DeadThread"
    ident = 0
    uuid = 0

    def __init__(self, uuid):
        self.ident = int(uuid)
        self.uuid = uuid

    def __repr__(self):
        return "DeadThread(%s)" % self.uuid

    def __eq__(self, other):
        if isinstance(other, DeadThread):
            return self.uuid == other.uuid
        return False

    def __hash__(self):
        return self.__repr__().__hash__()

    @classmethod
    def get(cls, uuid):
        return cls(uuid)


threading.Thread.parent = property(get_thread_parent)
threading.Thread.uuid = property(get_thread_uuid)

if is_module_patched("threading"):
    _REGISTER_GREENLETS = True
    import gevent.monkey
    gevent.monkey.saved['threading']['Thread'].parent = property(get_thread_parent)
    gevent.monkey.saved['threading']['Thread'].uuid = property(get_thread_uuid)
    IDENT_TO_UUID[main_thread_ident_before_patching] = get_thread_uuid()
    from weakref import WeakValueDictionary
    IDENT_TO_GREENLET = WeakValueDictionary()

    def iter_thread_frames():
        current_thread_ident = threading.current_thread().ident

        def fix(ident, frame):
            # if it is the current thread, we mustn't yield this very frame,
            # since it'll get detached from the stack once we exit this function
            return ident, (frame.f_back if ident == current_thread_ident else frame)

        main_thread_frame = None
        for ident, frame in sys._current_frames().items():
            if IDENT_TO_UUID.get(ident) == MAIN_UUID:
                main_thread_frame = frame
                # the MainThread should be shown in it's "greenlet" version
                continue
            yield fix(ident, frame)

        for thread in threading.enumerate():
            greenlet = IDENT_TO_GREENLET.get(thread.ident)
            if not greenlet:
                # some inbetween state, before greenlet started or after it died?...
                pass
            elif greenlet.gr_frame:
                yield fix(thread.ident, greenlet.gr_frame)
            else:
                # a thread with greenlet but without gr_frame will be fetched from sys._current_frames
                # If we switch to another greenlet by the time we get there we will get inconsistent dup of threads.
                # TODO - make best-effort attempt to show coherent thread dump
                yield fix(thread.ident, main_thread_frame)

else:
    def iter_thread_frames():
        yield from sys._current_frames().items()


def walk_frames(thread=None, *, across_threads=False):
    """
    Yields the stack frames of the current/specified thread.

    :param across_threads: yield frames of ancestor threads.
    Note that the parent thread might be off to other things, and not actually in the frame that spawned the thread at the tip.
    """

    if not thread:
        thread = threading.current_thread()

    frame = dict(iter_thread_frames()).get(thread.ident)

    while frame:

        yield frame

        frame = frame.f_back
        if not frame and across_threads and thread.parent:
            thread = thread.parent
            frame = dict(iter_thread_frames()).get(thread.ident)


this_module = import_module(__name__)
_BOOTSTRAPPERS = {threading, this_module}


def get_thread_trees(including_this=True):
    """
    Returns raw information about currently active threads, including their stack frames, and their immediate child-threads.

    The tree structure is maintained by special monkey-patching implemented by this module.
    """

    from .logging import THREAD_LOGGING_CONTEXT
    from .bunch import Bunch

    tree = {}
    dead_threads = set()
    contexts = {}
    stacks = {}

    def add_to_tree(thread):
        contexts[thread.ident] = THREAD_LOGGING_CONTEXT.flatten(thread.uuid)
        parent = get_thread_parent(thread)
        if isinstance(parent, DeadThread) and parent not in dead_threads:
            dead_threads.add(parent)
            add_to_tree(parent)
        tree.setdefault(parent, []).append(thread)

    for thread in threading.enumerate():
        add_to_tree(thread)

    current_ident = threading.current_thread().ident
    main_ident = threading.main_thread().ident

    for thread_ident, frame in iter_thread_frames():
        if not including_this and thread_ident == current_ident:
            formatted = "  <this frame>"
        else:
            # show the entire stack if it's this thread, don't skip ('after_module') anything
            show_all = thread_ident in (current_ident, main_ident)
            formatted = format_thread_stack(frame, skip_modules=[] if show_all else _BOOTSTRAPPERS) if frame else ''
        stacks[thread_ident] = formatted, time.time(), (id(frame) if frame else 0)

    def add_thread(parent_thread, parent):
        for thread in sorted(tree[parent_thread], key=lambda thread: thread.name):
            ident = thread.ident or 0
            stack, ts, frame_id = stacks.get(ident, ("", 0, 0))
            context = contexts.get(ident, {})
            context_line = ", ".join("%s: %s" % (k, context[k]) for k in "host context".split() if context.get(k))

            this = Bunch(
                name=thread.name,
                daemon="[D]" if getattr(thread, "daemon", False) else "",
                ident=ident,
                context_line="({})".format(context_line) if context_line else "",
                stack=stack,
                timestamp=ts,
                frame_id=frame_id,
                children=[],
                )
            parent.children.append(this)
            if thread in tree:
                add_thread(thread, this)
        return parent

    root = Bunch(children=[])
    add_thread(None, root)
    return root.children


def get_thread_stacks(including_this=True):
    """
    Returns an ``IndentableTextBuffer`` which renders the currently active threads.
    """

    stack_trees = get_thread_trees(including_this=including_this)

    def write_thread(branches):
        for branch in branches:
            with buff.indent('Thread{daemon}: {name} ({ident:X})', **branch):
                ts = time.strftime("%H:%M:%S", time.localtime(branch.timestamp))
                buff.write("{} Fid:{:X}  {}", ts, branch.frame_id, branch.context_line)
                for line in branch.stack.splitlines():
                    buff.write(line)
                write_thread(branch.children)

    from .humanize import IndentableTextBuffer
    buff = IndentableTextBuffer("Current Running Threads")
    write_thread(stack_trees)

    return buff


def watch_threads(interval, logger_name='threads'):
    """
    Starts a daemon thread that logs the active-threads to a ``threads`` logger, at the specified interval.
    The data is logged using the ``extra`` logging struct.
    It is recommended to configure the logger to use ``easypy.logging.YAMLFormatter``, so that the log can be
    easily parsed by other tools.
    """

    from easypy.resilience import resilient
    from easypy.concurrency import concurrent

    cmdline = " ".join(sys.argv)
    _logger = logging.getLogger(__name__)
    _threads_logger = logging.getLogger(logger_name)

    last_threads = set()

    @contextmanager
    def no_exceptions():
        try:
            yield
        except Exception:
            pass

    @no_exceptions()
    @resilient.warning
    def dump_threads():
        nonlocal last_threads

        with _logger.indented('getting thread tree', level=logging.DEBUG):
            trees = get_thread_trees(including_this=False)

        with _logger.indented('logging threads to yaml', level=logging.DEBUG):
            _threads_logger.debug("threads", extra=dict(cmdline=cmdline, tree=Bunch(children=trees).to_dict()))

        with _logger.indented('creating current thread set', level=logging.DEBUG):
            current_threads = set()
            for thread in threading.enumerate():
                if thread.ident:
                    current_threads.add(thread.ident)

        new_threads = current_threads - last_threads
        closed_threads = last_threads - current_threads
        stringify = lambda idents: ", ".join("%X" % ident for ident in idents)

        if new_threads:
            _logger.debug("WHITE<<NEW>> threads (%s): %s", len(new_threads), stringify(new_threads))
        if closed_threads:
            _logger.debug("threads terminated (%s): %s", len(closed_threads), stringify(closed_threads))
        _logger.debug("total threads: %s", len(current_threads))

        last_threads = current_threads

    thread = concurrent(dump_threads, threadname="ThreadWatch", loop=True, sleep=interval, real_thread_no_greenlet=True)
    thread.start()
    _logger.info("threads watcher started")


class ThreadContexts():
    """
    A structure for storing arbitrary data per thread.
    Unlike ``threading.local``, data stored here will be inherited into
    child-threads (threads spawned from threads)::

        TC = ThreadContexts()

        # (thread-a)
        with TC(my_data='b'):
            spawn_thread_b()

        # (thread-b)
        assert TC.my_data == 'b'
        with TC(my_data='c'):
            spawn_thread_c()
        assert TC.my_data == 'b'

        # (thread-c)
        assert TC.my_data == 'c'

    :param counters: attributes named here get incremented each time, instead of overwritten:
        ::
            TC = ThreadContexts(counters=('i', 'j'))
            assert TC.i == TC.j == 0

            with TC(i=1):
                assert TC.i == 1
                assert TC.j == 0

                with TC(i=1, j=1):
                    assert TC.i == 2
                    assert TC.j == 1


    :param stacks: attributes named here get pushed into a list, instead of overwritten:
        ::
            TC = ThreadContexts(stacks=('i', 'j'))
            with TC(i='a'):
                assert TC.i == ['a']
                assert TC.j == []
                with TC(i='i', j='j'):
                    assert TC.i == ['a', 'i']
                    assert TC.j == ['j']

    """

    def __init__(self, defaults={}, counters=None, stacks=None):
        self._context_data = WeakKeyDictionary()
        self._defaults = defaults.copy()
        self._counters = set(ilistify(counters or []))
        self._stacks = set(ilistify(stacks or []))

    def _get_context_data(self, thread_uuid=None, combined=False):
        if not thread_uuid:
            thread_uuid = get_thread_uuid()

        ctx = self._context_data.setdefault(thread_uuid, [])
        if not combined:
            return ctx

        parent_uuid = get_parent_uuid(thread_uuid)
        if parent_uuid:
            parent_ctx = self._get_context_data(parent_uuid, combined=True)
            ctx = deepcopy(parent_ctx) + ctx
        return ctx

    def get(self, k, default=None):
        """
        safely get the value of the specified key, in the current thread context
        """
        return self.flatten().get(k, default)

    def __getattr__(self, k):
        ret = self.get(k, AttributeError)
        if ret is AttributeError:
            raise AttributeError(k)
        return ret

    @contextmanager
    def __call__(self, kw=None, **kwargs):
        kw = dict(kw or {}, **kwargs)
        for v in kw.values():
            assert (v is None) or (isinstance(v, (str, int, float))), "Can't use %r as context vars" % v
        ctx = self._get_context_data()
        ctx.append(Bunch(kw))
        try:
            yield
        except Exception as exc:
            context = self.flatten()
            if context and not getattr(exc, "context", None):
                try:
                    exc.context = context
                except:
                    logging.warning("could not attach context to exception")
            raise
        finally:
            ctx.pop(-1)

    def flatten(self, thread_uuid=None):
        """
        return a flattened dict of all inherited context data for the current thread
        """
        stack = self._get_context_data(thread_uuid=thread_uuid, combined=True)
        concats = {k: self._defaults.get(k, []) for k in self._stacks}
        accums = {k: self._defaults.get(k, 0) for k in self._counters}
        extra = dict(self._defaults)
        for ctx in stack:
            for k in self._counters:
                accums[k] += ctx.get(k, 0)
            for k in self._stacks:
                if k in ctx:
                    concats[k].append(ctx[k])
            extra.update(ctx)
        extra.update(concats)
        extra.update(accums)
        return extra



def test_tokens():
    from easypy.tokens import AUTO, if_auto, MAX

    def foo(p=AUTO):
        return if_auto(p, 100)

    assert foo() == 100
    assert foo(5) == 5
    assert foo(MAX) == MAX

    assert MAX == "MAX"
    assert MAX == "<MAX>"
    assert MAX == "max"
    assert MAX == "<max>"

    d = {AUTO: AUTO, MAX: MAX}
    assert d[AUTO] == AUTO
    assert d[MAX] == MAX
    assert d['<MAX>'] is MAX
    assert 'AUTO' not in d



import threading

import logging
from easypy.concurrency import concurrent
from easypy.sync import RWLock
from easypy.bunch import Bunch


def test_rwlock():

    main_ctrl = threading.Event()
    reader_ctrl = threading.Event()
    writer_ctrl = threading.Event()
    lock = RWLock("test")

    state = Bunch(reading=False, writing=False)

    def read():
        logging.info("Before read")
        reader_ctrl.wait()
        reader_ctrl.clear()

        with lock:
            logging.info("Reading...")
            state.reading = True
            main_ctrl.set()
            reader_ctrl.wait()
            reader_ctrl.clear()
            state.reading = False
            logging.info("Done reading")

        logging.info("After read")

    def write():
        logging.info("Before write")
        writer_ctrl.wait()
        writer_ctrl.clear()

        with lock.exclusive():
            logging.info("Writing...")
            state.writing = True
            main_ctrl.set()
            writer_ctrl.wait()
            writer_ctrl.clear()
            state.writing = False
            logging.info("Done writing")

        logging.info("After write")
        main_ctrl.set()

    reader = concurrent(read, threadname='read')
    writer = concurrent(write, threadname='write')

    with reader, writer:
        assert not state.reading and not state.writing

        reader_ctrl.set()
        main_ctrl.wait()
        logging.info("lock acquired non-exclusively")
        main_ctrl.clear()
        assert state.reading and not state.writing

        writer_ctrl.set()
        logging.info("writer awaits exclusivity")
        with lock:
            assert state.reading and not state.writing

        reader_ctrl.set()
        main_ctrl.wait()
        main_ctrl.clear()
        logging.info("read lock released")
        assert not state.reading and state.writing

        writer_ctrl.set()
        main_ctrl.wait()
        main_ctrl.clear()
        logging.info("write lock released")
        assert not state.reading and not state.writing



import string
import random
from random import choice, sample

choose = choice


def random_nice_name(max_length=64, entropy=2, sep='-'):
    """Generates a nice random name from the dictionaries in words

    :param max_length: max length for the name.
    :type max_length: int, optional
    :param entropy: how unique th name will be, currently entropy - 1 adjectives are joined with one noun.
    :type entropy: int, optional
    :param sep: seperator between name parts.
    :type sep: str, optional

    :return: the generated name
    :rtype: str

    :raises ValueError: if ``param2`` is equal to ``param1``.
    """

    from .words import (adjectives, creatures)

    name = None
    entropy = max(entropy, 1)
    parts = (creatures, ) + (adjectives, ) * (entropy - 1)
    for _ in range(10):
        name_parts = [random.choice(p) for p in parts[::-1]]
        joined = sep.join(name_parts)
        if len(joined) <= max_length:
            name = joined
            break

    if not name:
        raise ValueError("Can't generate name under these conditions")

    return name


def random_string(length, charset=string.printable):
    return ''.join(random.choice(charset) for i in range(length))


def random_filename(length=(3, 50)):
    if hasattr(length, "__iter__"):
        mn, mx = length
        length = random.randrange(mn, mx+1)  # randrange does not include upper bound
    return random_string(length, charset=string.ascii_letters)


def random_buf(size):
    assert size < 5 * 2**20, "This is too big for a buffer (%s)" % size
    return random_string(size).encode("latin-1")


def perchance(probabilty):
    return random.random() <= probabilty



def test_colors():
    from easypy.colors import Colorized, uncolored, colorize, register_colorizers

    register_colorizers(bad=("red", "blue"))

    opts = {
        str(Colorized("RED(BLUE)<<XXX>>")),
        str(Colorized("RED(BLUE)@[XXX]@")),
        str(Colorized("RED(BLUE)@{XXX}@")),

        str(colorize("RED(BLUE)<<XXX>>")),
        str(colorize("RED(BLUE)@[XXX]@")),
        str(colorize("RED(BLUE)@{XXX}@")),

        str(colorize("BAD<<XXX>>")),
        str(colorize("BAD@[XXX]@")),
        str(colorize("BAD@{XXX}@")),
    }

    assert len(opts) == 1
    [ret] = opts
    assert ret == "\x1b[1;44;31mXXX\x1b[0m"
    assert uncolored(ret) == "XXX"


def test_empty_markup():
    from easypy.colors import Colorized, colorize

    opts = {
        colorize("BLUE<<>>"),
        colorize("BLUE@[]@"),
        colorize("BLUE@{}@"),
        str(Colorized("BLUE<<>>")),
        str(Colorized("BLUE@[]@")),
        str(Colorized("BLUE@{}@")),
    }

    assert len(opts) == 1
    [ret] = opts

    assert ret == ""



from contextlib import contextmanager
import inspect
from functools import wraps

from easypy.properties import cached_property


class Fixture(object):
    def __init__(self, name, function, cached=True):
        self.name = name
        self._function = function
        self._cached = cached

    def __repr__(self):
        return 'Fixture %s' % self.name

    @cached_property
    def _signature(self):
        return inspect.signature(self._function)

    @cached_property
    def dependencies(self):
        return self._signature.parameters.keys()

    def invoke(self, fixtures_assembly):
        kwargs = {dependency: fixtures_assembly.resolve_fixture(dependency) for dependency in self.dependencies}
        return self._function(**kwargs)


class FixturesNamespace(object):
    def __init__(self):
        self.fixtures = {}

    def register(self, func=None, *, cached=True):
        @wraps(func)
        def inner(func):
            assert func.__name__ not in self.fixtures, 'Fixture %s already exists' % func.__name__
            fixture = Fixture(func.__name__, func, cached=cached)
            self.fixtures[fixture.name] = fixture

        if func is None:
            return inner
        else:
            return inner(func)

    def assemble(self, **manual_fixture_values):
        assembly = FixturesAssembly(self)
        for k, v in manual_fixture_values.items():
            assembly._set_fixture_value(k, v)
        return assembly

    def get(self, name):
        return self.fixtures[name.split('__')[0]]


class FixturesAssembly(object):
    def __init__(self, fixtures_namespace):
        self._fixtures_namespace = fixtures_namespace
        self._fixture_values = {}
        self.__being_added = set()

    @contextmanager
    def __adding(self, name):
        assert name not in self.__being_added, \
            'Circular dependency detected while invoking fixture %s. Currently resolving %s' % (name, self.__being_added)
        self.__being_added.add(name)
        try:
            yield
        finally:
            self.__being_added.remove(name)

    def _set_fixture_value(self, name, value):
        assert name not in self._fixture_values, 'Value for fixture %s already set' % (name,)
        self._fixture_values[name] = value

    def resolve_fixture(self, name):
        try:
            return self._fixture_values[name]
        except KeyError:
            pass

        with self.__adding(name):
            fixture = self._fixtures_namespace.get(name)
            fixture_value = fixture.invoke(self)

        if fixture._cached:
            self._set_fixture_value(name, fixture_value)

        return fixture_value

